from typing import List, Union

import numpy as np

# TODO
# Version number changed automatically using python-semantic-release
__version__ = "0.1.0"

# ---------------------------------
# Types
# ---------------------------------
REAL_TYPE = Union[float, int]
REAL_VECT_TYPE = Union[float, int, np.ndarray, List[REAL_TYPE]]

# ---------------------------------
# Plotting
# ---------------------------------
RANGE_FACECOLOR = "blue"
PLOT_STYLE = "seaborn-white"
DPI = 400



"""Target functions for testing."""

import numpy as np


def lldeh_1d(x: np.ndarray, a: float) -> np.ndarray:
    """
    An interesting 1D non-linear function.

    Source: https://www.reddit.com/r/math/comments/5mwr9g/what_are_your_favourite_or_most_interesting/  # noqa E301
    Args:
        x:
        a: parameter to control the shape, interesting variants: a=0, a=2.1

    Returns:

    """
    g = np.tan(0.5 * a)
    return 2 * x**2 / (x**2 + g * x + 1)


def peaks_2d(x: np.ndarray) -> np.ndarray:
    # https://nl.mathworks.com/help/matlab/ref/peaks.html
    x = np.atleast_2d(x)
    x1 = x[:, 0]
    x2 = x[:, 1]

    return (
        3 * (1 - x1) ** 2 * np.exp(-(x1**2) - (x2 + 1) ** 2)
        - 10 * (x1 / 5 - x1**3 - x2**5) * np.exp(-(x1**2) - x2**2)
        - 1 / 3 * np.exp(-((x1 + 1) ** 2) - x2**2)
    )


def peaks_2d_multivariate(x: np.ndarray) -> np.ndarray:
    # https://nl.mathworks.com/help/matlab/ref/peaks.html
    x = np.atleast_2d(x)
    x1 = x[:, 0]
    x2 = x[:, 1]

    result = (
        3 * (1 - x1) ** 2 * np.exp(-(x1**2) - (x2 + 1) ** 2)
        - 10 * (x1 / 5 - x1**3 - x2**5) * np.exp(-(x1**2) - x2**2)
        - 1 / 3 * np.exp(-((x1 + 1) ** 2) - x2**2)
    )
    return np.stack((result, result**2), -1)


def ackley_nd(
    x_mx: np.ndarray, a: float = 20.0, b: float = 0.2, c: float = 2 * np.pi
) -> np.ndarray:
    """
    n-dimensional Ackley function based on https://www.sfu.ca/~ssurjano/ackley.html.

    Args:
        x_mx:
            Independent variable values, [x1, x2, ..., xd], vectorized input is
            supported (provide them along the row dimension).
        a:
            Constant/parameter in the function.
        b:
            Constant/parameter in the function.
        c:
            Constant/parameter in the function.

    Returns:
        Dependent variable values.
    """
    x_mx = np.atleast_2d(x_mx)
    t1 = -a * np.exp(-b * np.sqrt(np.mean(x_mx**2, axis=1)))
    t2 = -np.exp(np.mean(np.cos(c * x_mx), axis=1))

    return t1 + t2 + a + np.exp(1)


def six_hump_camel_2D(X: np.ndarray) -> np.ndarray:
    x = X[:, 0]
    y = X[:, 1]

    x2 = np.power(x, 2)
    x4 = np.power(x, 4)
    y2 = np.power(y, 2)

    return ((4.0 - 2.1 * x2 + (x4 / 3.0)) * x2) + (x * y) + ((-4.0 + 4.0 * y2) * y2)


def six_hump_camel_2D_2input(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    x2 = np.power(x, 2)
    x4 = np.power(x, 4)
    y2 = np.power(y, 2)

    return ((4.0 - 2.1 * x2 + (x4 / 3.0)) * x2) + (x * y) + ((-4.0 + 4.0 * y2) * y2)


def forrester_1d(x: np.ndarray) -> np.ndarray:
    term_a = 6 * x - 2
    term_b = 12 * x - 4
    return np.power(term_a, 2) * np.sin(term_b)


def bohachevsky_2D(X: np.ndarray) -> np.ndarray:
    x1 = X[:, 0]
    x2 = X[:, 1]

    t1 = np.power(x1, 2)
    t2 = 2 * np.power(x2, 2)
    t3 = -0.3 * np.cos(3 * np.pi * x1)
    t4 = -0.4 * np.cos(4 * np.pi * x2)

    return t1 + t2 + t3 + t4 + 0.7


def shekel(X: np.ndarray) -> np.ndarray:
    xx = [X[:, 0], X[:, 1], X[:, 2], X[:, 3]]
    m = 10
    b = 0.1 * np.array([1, 2, 2, 4, 4, 6, 3, 7, 5, 5])
    C = np.array(
        [
            [4.0, 1.0, 8.0, 6.0, 3.0, 2.0, 5.0, 8.0, 6.0, 7.0],
            [4.0, 1.0, 8.0, 6.0, 7.0, 9.0, 3.0, 1.0, 2.0, 3.6],
            [4.0, 1.0, 8.0, 6.0, 3.0, 2.0, 5.0, 8.0, 6.0, 7.0],
            [4.0, 1.0, 8.0, 6.0, 7.0, 9.0, 3.0, 1.0, 2.0, 3.6],
        ]
    )

    outer = 0
    for i in range(0, m):
        b_i = b[i]
        inner = 0
        for j in range(0, 4):
            x_j = xx[j]
            C_ji = C[j, i]
            inner = inner + np.power((x_j - C_ji), 2)
        outer = outer + 1 / (inner + b_i)

    return -1.0 * outer


def hartmann(X: np.ndarray) -> np.ndarray:
    n = X.shape[0]
    results = []
    for i in range(n):
        alpha = np.array([1.0, 1.2, 3.0, 3.2])
        A = np.array(
            [
                [10, 3, 17, 3.5, 1.7, 8],
                [0.05, 10, 17, 0.1, 8, 14],
                [3, 3.5, 1.7, 10, 17, 8],
                [17, 8, 0.05, 10, 0.1, 14],
            ]
        )
        P = 1e-4 * np.array(
            [
                [1312, 1696, 5569, 124, 8283, 5886],
                [2329, 4135, 8307, 3736, 1004, 9991],
                [2348, 1451, 3522, 2883, 3047, 6650],
                [4047, 8828, 8732, 5743, 1091, 381],
            ]
        )

        outer = 0
        for ii in range(4):
            inner = 0
            for jj in range(6):
                xj = X[i, jj]
                Aij = A[ii, jj]
                Pij = P[ii, jj]
                inner = inner + Aij * (xj - Pij) ** 2

            new = alpha[ii] * np.exp(-inner)
            outer = outer + new
        results.append(-(2.58 + outer) / 1.94)

    return np.asarray(results)


def stybtang(X: np.ndarray) -> np.ndarray:
    # https://www.sfu.ca/~ssurjano/stybtang.html
    n = X.shape[0]
    d = X.shape[1]
    res = np.zeros(n)
    for i in range(n):
        _sum = 0.0
        for j in range(d):
            _sum = np.power(X[i, j], 4) - 16.0 * np.power(X[i, j], 2) + 5.0 * X[i, j]
        res[i] = 0.5 * _sum

    return res



"""Fuzzy Lola-Vornoi adaptive design strategy for global surrogate modelling.

The algorithm is proposed and described in this paper:
[1] Crombecq, Karel, et al. (2011) A novel hybrid sequential design strategy for global
surrogate modeling of computer experiments. SIAM Journal on Scientific Computing 33.4
(2011): 1948-1974.

[2] van der Herten, J., Couckuyt, I., Deschrijver, D., & Dhaene, T. (2015).
A fuzzy hybrid sequential design strategy for global surrogate modeling of
 high-dimensional computer experiments.
SIAM Journal on Scientific Computing, 37(2), A1020-A1039.
"""
import os
import pickle
import time
from typing import Callable, Tuple

import numpy as np
import skfuzzy as fuzz
from loguru import logger
from scipy.spatial.distance import cdist, pdist, squareform
from skfuzzy import control as ctrl

from harlow.sampling.sampling_baseclass import Sampler
from harlow.surrogating.surrogate_model import Surrogate
from harlow.utils.helper_functions import evaluate, latin_hypercube_sampling
from harlow.utils.log_writer import write_scores, write_timer
from harlow.utils.metrics import rmse


# -----------------------------------------------------
# USER FACING API (class)
# -----------------------------------------------------
class FuzzyLolaVoronoi(Sampler):
    def __init__(
        self,
        target_function: Callable[[np.ndarray], np.ndarray],
        surrogate_model: Surrogate,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = rmse,
        logging_metrics: list = None,
        verbose: bool = False,
        run_name: str = None,
        save_dir: str = "",
    ):

        super(FuzzyLolaVoronoi, self).__init__(
            target_function,
            surrogate_model,
            domain_lower_bound,
            domain_upper_bound,
            fit_points_x,
            fit_points_y,
            test_points_x,
            test_points_y,
            evaluation_metric,
            logging_metrics,
            verbose,
            run_name,
            save_dir,
        )

        self.dim_in = len(domain_lower_bound)
        self.dim_out = None if self.fit_points_x is None else self.fit_points_y.shape[1]

        self.surrogate_model = surrogate_model()

        if self.dim_out > 1:
            self.multiresponse_sampling = True
            if not self.surrogate_model.is_multioutput:
                raise ValueError(
                    "Multiresponse target requires \
                                 multiresponse surrogate"
                )

    def sample(
        self,
        n_initial_points: int = 20,
        n_new_points_per_iteration: int = 1,
        stopping_criterium: float = None,
        max_n_iterations: int = 5000,
    ):
        """TODO: allow for providing starting points"""
        # ..........................................
        # Initialize
        # ..........................................
        target_function = self.target_function
        domain_lower_bound = self.domain_lower_bound
        domain_upper_bound = self.domain_upper_bound

        if n_initial_points is None:
            n_initial_points = 5 * self.dim_in

        # ..........................................
        # Initial sample of points
        # ..........................................
        gen_start_time = time.time()
        if self.fit_points_x is None:
            # latin hypercube sampling to get the initial sample of points
            points_x = latin_hypercube_sampling(
                n_sample=n_initial_points,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
            )
            # evaluate the target function
            points_y = target_function(points_x)

            self.fit_points_x = points_x
            self.fit_points_y = points_y
            self.dim_out = self.fit_points_x.shape[0]
        else:
            points_x = self.fit_points_x
            points_y = self.fit_points_y

        self.step_gen_time.append(time.time() - gen_start_time)

        # fit the surrogate model
        start_time = time.time()

        self.surrogate_model.fit(points_x, points_y)
        logger.info(
            f"Fitted the first surrogate model in {time.time() - start_time} sec."
        )

        predicted_y = self.surrogate_model.predict(self.test_points_x, as_array=True)

        score = evaluate(self.logging_metrics, self.test_points_y, predicted_y)
        self.step_x.append(points_x)
        self.step_y.append(points_y)
        self.step_score.append(score[self.evaluation_metric.__name__])
        self.step_iter.append(0)
        self.step_fit_time.append(time.time() - start_time)

        # ..........................................
        # Iterative improvement (adaptive stage)
        # ..........................................

        for ii in range(max_n_iterations):
            logger.info(
                f"Started adaptive iteration step: {ii+1} (max steps:"
                f" {max_n_iterations})."
            )

            start_time = time.time()
            new_points_x = best_new_points(
                points_x=points_x,
                points_y=points_y,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
                n_new_point=n_new_points_per_iteration,
                dim_in=self.dim_in,
            )
            self.step_gen_time.append(time.time() - start_time)
            logger.info(
                f"Found the next best {n_new_points_per_iteration} point(s) in "
                f"{time.time() - start_time} sec."
            )

            # evaluate the target function
            new_points_y = target_function(new_points_x)

            # add the new points to the old ones
            points_x = np.vstack((points_x, new_points_x))
            points_y = np.vstack((points_y, new_points_y))

            # refit the surrogate
            start_time = time.time()
            self.surrogate_model.update(new_points_x, new_points_y.ravel())
            self.step_fit_time.append(time.time() - start_time)
            logger.info(
                f"Fitted a new surrogate model in {time.time() - start_time} sec."
            )
            #
            self.fit_points_x = points_x
            self.fit_points_y = points_y

            # Re-evaluate the surrogate model.
            predicted_y = self.surrogate_model.predict(
                self.test_points_x, as_array=True
            )
            score = evaluate(self.logging_metrics, self.test_points_y, predicted_y)

            self.step_x.append(points_x)
            self.step_y.append(points_y)
            self.step_score.append(score)
            self.step_iter.append(ii + 1)
            timing_dict = {
                "Gen time": self.step_gen_time[ii + 1],
                "Fit time": self.step_fit_time[ii + 1],
            }
            write_scores(self.writer, score, ii + 1)
            write_timer(self.writer, timing_dict, ii + 1)

            # Currently use RMSE for convergence
            self.score = score[self.evaluation_metric.__name__]
            self.iterations = ii
            # Save model every 200 iterations
            if self.iterations % 200 == 0:
                save_name = self.run_name + "_{}_iters.pkl".format(self.iterations)
                save_path = os.path.join(self.save_dir, save_name)
                with open(save_path, "wb") as file:
                    pickle.dump(self.surrogate_model, file)

            if self.score <= stopping_criterium or ii >= max_n_iterations:
                logger.info(f"Algorithm converged in {ii} iterations")
                # Save model if converged
                with open(save_path, "wb") as file:
                    pickle.dump(self.surrogate_model, file)
                break

        self.writer.close()
        return self.fit_points_x, self.fit_points_y


# -----------------------------------------------------
# SUPPORTING FUNCTIONS
# -----------------------------------------------------
def best_new_points(
    points_x: np.ndarray,
    points_y: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_new_point: int,
    dim_in: int,
) -> np.ndarray:
    """
    Hybrid Sequential Strategy - Alg. (2) [2].
    Combines an exploitation (FLOLA) and an exploration (Voronoi) score
    and selects a new candindate sample in the neighborhood of the N_new
    highest ranked samples

    :param points_x: Nxd-dimensional input data points
    :param points_y: Nxd-dimensional target data points
    :param domain_lower_bound: The lower bound of the function's space
    :param domain_upper_bound: The upper bound of the function's space
    :param n_new_point: The number of new points we wish to sample
    :return new_reference_point_x: The new highly-ranked samples in the neighborhood
    """

    # shape the input if not in the right shape
    points_x = points_x.reshape((-1, dim_in))
    points_y = points_y.reshape((-1, 1))
    # Calculate distance matrix P
    distance_matrix = calculate_distance_matrix(points_x, dim_in)
    # Calculate the V for every Pr
    (
        relative_volumes,
        random_points,
        distance_mx,
        closest_indicator_mx,
    ) = voronoi_volume_estimate(
        points=points_x,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
    )

    neighborhoods_scores = best_neighbourhoods(
        points_x, points_y, distance_matrix, relative_volumes
    )

    # TODO: check np.partition as an alternative
    idxs_new_neighbor = np.argsort(-neighborhoods_scores)[:n_new_point]

    new_reference_points_x = np.empty((n_new_point, dim_in))
    for ii, idx_new_neighbor in enumerate(idxs_new_neighbor):
        # all the distances to reference point whose neighborhood will get a new
        # point where the target function is evaluated
        distances_in_neighbor = distance_mx[:, idx_new_neighbor]

        # consider only those points that are within the voronoi cell of the
        # reference point
        idx_mask_in_neighbor = closest_indicator_mx[:, idx_new_neighbor]

        # to avoid selecting points that are not in the neighborhood
        distances_in_neighbor[~idx_mask_in_neighbor] = -1

        # largest distance within the same voronoi cell
        idx_max_distance = np.argmax(distances_in_neighbor)
        new_reference_point_x = random_points[idx_max_distance]
        new_reference_points_x[ii, :] = new_reference_point_x

    return new_reference_points_x


def best_neighbourhoods(
    points_x: np.ndarray,
    points_y: np.ndarray,
    distance_matrix: np.ndarray,
    volume_estimate: np.ndarray,
) -> np.ndarray:
    """
    Exploitation Aglorithm - Alg. (1) [2].
    Computes a score fir all points p in P_r, indicating the nonlinearity
    of the region surrounding p. New samples are chosen in the neighborhood
    of the N_new highest ranked samples.

    :param points_x: Nxd-dimensional input data points
    :param points_y: Nxd-dimensional target data points
    :param distance_matrix: NxN distance matrix
    :volume_estimate: N-dimensional Voronoi volume estimate
    :returns: H_fuzz: N-dimensional Hybrid score
    """

    n_point, n_dim = points_x.shape
    # print('X shapes', points_x.shape)
    # print('Y shapes', points_y.shape)
    nonlinear_score, H_fuzzy = np.empty(n_point), np.empty(n_point)
    # Init FIS S
    # It should be done before the for loop BUT
    # requires adhesion MAX value, that is computed inside the LOOP
    # TODO CHeck it later.
    # FIS = init_FIS()
    K = 4 * n_dim
    if K >= n_point - 1:
        K = n_point - 1

    for ii in range(n_point):
        alpha = calculate_alpha(ii, K, distance_matrix)
        # gets the neighbours of Pr as indices
        neighbors_idx = get_neighbourhood(ii, distance_matrix, alpha, n_dim)
        # print("Neighbors index", neighbors_idx, '\n', neighbors_idx.shape)
        neighbors_coords = points_x[neighbors_idx, :]
        # print("Neighbor coords ", neighbors_coords, '\n', neighbors_coords.shape)
        adhesion, cohesion = get_adhesion_cohesion(ii, neighbors_idx, distance_matrix)
        # print("ADH & COH ", adhesion, adhesion.shape, cohesion, cohesion.shape)
        FIS = init_FIS(neighbors_coords, adhesion)
        w = assign_weights(neighbors_coords, cohesion, adhesion, FIS)
        # print('Weights', w)
        grad = flola_gradient_estimate(
            points_x[ii, :],
            points_y[ii],
            points_x[neighbors_idx, :],
            points_y[neighbors_idx, :],
            w,
        )
        # print('grad', grad, grad.shape)
        # E_fuzzy(P_r) Eq. (3.2) [2]
        nonlinear_score[ii] = nonlinearity_measure(
            points_x[ii, :],
            points_y[ii, :],
            grad,
            points_x[neighbors_idx, :],
            points_y[neighbors_idx, :],
        )
        # print('Nonlinear score', nonlinear_score[ii])

    # H_fuzzy(P_r) Eq.(5.1) [2]
    H_fuzzy = flola_voronoi_score(nonlinear_score, volume_estimate)
    # print(H_fuzzy, H_fuzzy.shape)
    return H_fuzzy


def init_FIS(data_points: np.ndarray, adhesion: np.ndarray):
    """
    Initialize the Fuzzy Inference System S; Section 4 [2].

    :param data_points: Nxd-dimensional input vector of N data points with d dimensions
    :param cohesion: The N-dimensional cohesion values of the neighbors of P_r
    :param adhesion: The N-dimensional adhesion values of the neighbors of P_r
    :return flola_sim: The Fuzzy Inference System S
    """

    a = np.sort(data_points, axis=None)
    # Different ways to define the function's ranges.
    # Sparse = Faster / Dense = Slower
    # a = np.arange(np.min(data_points), np.max(data_points), 0.1)
    # a_w = np.linspace(0, 1, 100, endpoint=True)
    # a_w = np.linspace(0, 1, 5, endpoint=True)
    a_w = np.arange(0, 1, 0.1)
    A_max = np.max(adhesion)
    coh = ctrl.Antecedent(a, "cohesion")
    adh = ctrl.Antecedent(a, "adhesion")
    wei = ctrl.Consequent(a_w, "weight")

    # Define the membership functions & populate the space
    coh["high"] = coh_high(coh.universe)
    adh["low"] = adh_low(adh.universe, A_max)
    adh["high"] = adh_high(adh.universe, A_max)

    # construct triangular output/weight member functions.
    wei["low"] = fuzz.trimf(wei.universe, [0, 0, 0.13])
    wei["average"] = fuzz.trimf(wei.universe, [0.16, 0.33, 0.5])
    wei["high"] = fuzz.trimf(wei.universe, [0.6, 1.01, 1.01])
    # Define the rules of the System
    rule1 = ctrl.Rule(coh["high"] & adh["low"], wei["high"])
    rule2 = ctrl.Rule(coh["high"] & adh["high"], wei["average"])
    rule3 = ctrl.Rule(~coh["high"] & adh["low"], wei["average"])
    rule4 = ctrl.Rule(~coh["high"] & adh["high"], wei["low"])
    # Setup the FIS
    FLOLA_ctrl = ctrl.ControlSystem([rule1, rule2, rule3, rule4])
    flola_sim = ctrl.ControlSystemSimulation(FLOLA_ctrl)

    return flola_sim


# TODO time it if needed
def assign_weights(
    data_points: np.ndarray,
    cohesion: np.ndarray,
    adhesion: np.ndarray,
    flola_sim,
) -> np.ndarray:
    """
    Weight calculation of data_points based on cohesion & adhesion values between
    the data points; Section 4 [2].

    :param data_points: Nxd-dimensional input vector of N data points with d dimensions
    :param cohesion: The N-dimensional cohesion values of the neighbors of P_r
    :param adhesion: The N-dimensional adhesion values of the neighbors of P_r
    :param flola_sim: The Fuzzy Inference System S
    :return weight: The N-dimensional weight values returned by the evaluation of FIS S
    """

    S = flola_sim
    dims = data_points.shape
    weights = np.zeros(dims[0])

    # Get the weights for each neighbor
    for i in range(dims[0]):
        S.input["cohesion"] = cohesion[i]
        S.input["adhesion"] = adhesion[i]
        S.compute()
        weights[i] = S.output["weight"]

    return weights


def coh_high(x, s_c=0.3):
    """
    Cohesion High membership function
    """
    return 1 / (1 + np.exp(-s_c * x))


def adh_high(x, A_max, s_ah=0.3):
    """
    Adhesion High membership function
    """
    return np.exp(((-(x**2)) / 2 * ((A_max * s_ah) ** 2)))


def adh_low(x, A_max, s_al=0.27):
    """
    Adhesion Low membership function
    """
    res = (-((x - A_max) ** 2)) / 2 * ((A_max * s_al) ** 2)
    return np.exp(res)


def get_adhesion_cohesion(
    Pr_index: int, P_neigbors_idxs: np.ndarray, distance_matix: np.ndarray
) -> Tuple[np.array, np.array]:
    """
    Calculation of adhesion and cohesion values for the neighbors of P_r.
    According to Eq. (4.1) and (4.2) [2].

    :param Pr_index: Index of reference point
    :param P_neigbors_idxs: neighbour indexes for point Pr
    :param distance_matix: The precalculated distance matrix for all p in P
    :return: Adhesion and Cohesion arrays for N
    """

    C = distance_matix[Pr_index, P_neigbors_idxs]
    A = np.zeros((len(P_neigbors_idxs)))

    for i in range(len(P_neigbors_idxs)):
        neighbor_dists = distance_matix[P_neigbors_idxs[i], :]
        r = np.delete(neighbor_dists, P_neigbors_idxs[i])
        A[i] = np.min(r)

    return A, C


def get_neighbourhood(
    Pr_idx: int, distance_matrix: np.ndarray, alpha: float, n_dim: int
) -> np.array:
    """
    Calculate the neighbourhood for point Pr.
    Eq. (3.3) [2].

    :param Pr_idx: Index of reference point
    :param distance_matrix: The NxN-dimensional precalculated distance matrix
    for all p in P
    :param alpha: Distance parameter Eq. (3.4) [2]
    :param n_dim: Number of dimensions
    :return neighbors_idx: N-dimensional array with neighbour indexes for
    point Pr
    """

    distances_prIdx = distance_matrix[Pr_idx, :]
    neighbors_idx = np.where(distances_prIdx < alpha)[0]
    neighbors_idx = neighbors_idx[neighbors_idx != Pr_idx]

    # If number of neighbors is smaller than ndim Eq. (3.1) will be undertermined.
    # Then, take ndim nearest neighbors.
    if len(neighbors_idx) < n_dim:
        nearest_ndim_idx = np.argpartition(np.delete(distances_prIdx, Pr_idx), n_dim)
        neighbors_idx = distances_prIdx[nearest_ndim_idx[:n_dim]]

    return neighbors_idx


def calculate_alpha(Pr_idx: int, K: int, distance_matrix: np.ndarray) -> np.ndarray:
    """
    Calculate distance parameter alpha.
    Eq. (3.4) [2].

    :param Pr_idx: Index of reference point
    :param K: Parameter K = 4d
    :param distance_matrix: The precalculated distance matrix for all p in P
    :return: alpha as float
    """

    distances_prIdx = distance_matrix[Pr_idx, :]
    distances_wo_prIdx = np.delete(distances_prIdx, Pr_idx)
    nearest_k_vals = np.sort(distances_wo_prIdx)[:K]

    return np.sum(nearest_k_vals) * (2 / K)


def calculate_distance_matrix(
    points_x: np.ndarray, n_dim: int, fractional: bool = False
) -> np.ndarray:
    """
    This function calculates the distance matrix for the points P.
    Instead of calculating the distance for alpha, A and
    C for Pr in the loop, this function is meant to be called once,
    prior to the main for loop iterating over all Pr.

    :param points_x: The set of point P
    :return: A distance matrix for P
    """

    if not fractional:
        return squareform(pdist(points_x, "euclidean"))
    else:
        return squareform(pdist(points_x, "minkowsky", p=n_dim))


def flola_voronoi_score(
    nonlinearity_measures: float, relative_volumes: np.ndarray
) -> np.ndarray:
    """Eq.(5.1) of [2]."""
    return relative_volumes + nonlinearity_measures / np.sum(nonlinearity_measures)


def nonlinearity_measure(
    reference_point_x: np.ndarray,
    reference_point_y: float,
    reference_point_gradient: np.ndarray,
    neighbor_points_x: np.ndarray,
    neighbor_points_y: np.ndarray,
) -> float:
    """Eq.(4.9) of [1]."""
    e = np.sum(
        np.abs(
            neighbor_points_y
            - (
                reference_point_y
                + reference_point_gradient * (neighbor_points_x - reference_point_x)
            )
        )
    )
    return float(e)


def voronoi_volume_estimate(
    points: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_simulation: int = None,
    random_points: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Estimate the relative volume of the Voronoi tessellation of `points` bounded by
    `domain_lower_bound` and `domain_upper_bound`.

    The algorithm is described in section 3 of [1].

    Args:
        points:
            n_point x n_dim.
        domain_lower_bound:
            n_dim.
        domain_upper_bound:
            n_dim
        n_simulation:
            Number of random points used to estimate the relative volumes. If
            `random_points` is provided then this argument is ignored.
        random_points:

    Returns:
        Relative volumes in the same order as `points`.
    """
    # dimensions are not checked
    if n_simulation is None:
        n_simulation = 1000 * points.shape[0]

    if random_points is None:
        n_dim = len(domain_lower_bound)
        random_points = domain_lower_bound + np.random.rand(n_simulation, n_dim) * (
            domain_upper_bound - domain_lower_bound
        )

    # all relevant distances, n_simulation x n_point
    distance_mx = cdist(random_points, points, metric="euclidean")

    # index of the closest `point` to each `random_point`
    col_idx_min = np.argmin(distance_mx, axis=1)
    row_idx_min = np.arange(distance_mx.shape[0])

    # indicator matrix to count the number of `random_points` closest to each `points`
    closest_indicator_mx = np.zeros(distance_mx.shape, dtype=bool)
    # place a one (True) in the indicator matrix if the element (distance) is a closest
    # distance
    closest_indicator_mx[row_idx_min, col_idx_min] = True

    # count the closest `random_points` for each `points`
    closest_counts = np.sum(closest_indicator_mx, axis=0)

    # relative volume estimate
    relative_volumes = closest_counts / n_simulation

    return relative_volumes, random_points, distance_mx, closest_indicator_mx


def flola_gradient_estimate(
    reference_point_x: np.ndarray,
    reference_point_y: np.ndarray,
    neighbor_points_x: np.ndarray,
    neighbor_points_y: np.ndarray,
    weights: np.ndarray,
) -> np.ndarray:
    """
    Estimate the gradient at `reference_point` (P_r) by fitting a hyperplane to
    `neighbor_points` in a least-square sense. A hyperplane that goes exactly
    through the `reference_point`. Eq.(3.2) [2] wrt to weights for all neighbors of P_r
    obtained from solving FIS S.

    :param reference_point_x: The d-dimensional reference sample P_r
    :param reference_point y: -//-
    :param neigbor_points_x: The m x d-dimensional neighbors of P_r
    :param neigbor_points_y: -//-
    :param weights: The weights of every neighbor of P_r
    :return gradient: The d-dimensional gradient at P_r
    """

    reference_point_x = reference_point_x.reshape((1, -1))
    n_neighbors, n_dims = neighbor_points_x.shape
    # to ensure that we hyperplane goes through `reference_point`
    neighbor_points_x_diff = neighbor_points_x - reference_point_x
    neighbor_points_y_diff = neighbor_points_y - reference_point_y
    # Solve Weighted Least Squares
    # Least-Square fit of the hyperplane, the gradient is the hyperplane coefficient
    Aw = neighbor_points_x_diff * np.sqrt(weights[:, np.newaxis])
    Bw = neighbor_points_y_diff * np.sqrt(weights)
    gradient = np.linalg.lstsq(Aw, Bw, rcond=None)[0].reshape(n_neighbors, n_dims)

    return gradient



"""Cross Validation Voronoi adaptive design strategy for global surrogate
modelling.

Note that this

The algorithm is proposed and described in this paper:
[1] Xu, Shengli, et al. A robust error-pursuing sequential sampling approach
 for global metamodeling based on voronoi diagram and cross validation.
 Journal of Mechanical Design 136.7 (2014): 071009

[2] Liu H, Xu S, Wang X, Yang S, Meng J. A (1996).
A multi-response adaptive sampling approach for global metamodeling
Proceedings of the Institution of Mechanical Engineers Part C Journal of
Mechanical Engineering Science 1989-1996 (vols 203-210)

[3] Kaminsky A.L., Wang Y, Pant K, Hashii W.N., Atachbarian A. (2020) An
efficient batch k-fold cross-validation voronoi adaptive sampling technique
for global surrogate modelling. Journal of Mechanical Design 143
 """

import os
import pickle
import time
from typing import Callable, Tuple

import numpy as np
from loguru import logger
from scipy.spatial.distance import cdist
from sklearn.model_selection import KFold

from harlow.sampling.sampling_baseclass import Sampler
from harlow.surrogating.surrogate_model import Surrogate
from harlow.utils.helper_functions import (
    evaluate_modellist,
    latin_hypercube_sampling,
    normalized_response,
)
from harlow.utils.log_writer import write_scores, write_timer
from harlow.utils.metrics import nrmse, rmse


# -----------------------------------------------------
# USER FACING API (class)
# -----------------------------------------------------
class CVVoronoi(Sampler):
    def __init__(
        self,
        target_function: Callable[[np.ndarray], np.ndarray],
        surrogate_model: Surrogate,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = rmse,
        logging_metrics: list = None,
        verbose: bool = False,
        run_name: str = None,
        save_dir: str = "",
    ):

        super(CVVoronoi, self).__init__(
            target_function,
            surrogate_model,
            domain_lower_bound,
            domain_upper_bound,
            fit_points_x,
            fit_points_y,
            test_points_x,
            test_points_y,
            evaluation_metric,
            logging_metrics,
            verbose,
            run_name,
            save_dir,
        )

        self.surrogates = []

    def sample(
        self,
        n_initial_points: int = 20,
        n_new_points_per_iteration: int = 1,
        stopping_criterium: float = 0.05,
        max_n_iterations: int = 5000,
    ):
        """TODO: allow for providing starting points"""
        # ..........................................
        # Initialize
        # ..........................................
        target_function = self.target_function
        domain_lower_bound = self.domain_lower_bound
        domain_upper_bound = self.domain_upper_bound
        n_dim = len(domain_lower_bound)

        """This is to support multi-output"""
        dim_out = self.fit_points_y.shape[1]

        if n_initial_points is None:
            n_initial_points = 5 * n_dim

        # ..........................................
        # Initial sample of points
        # ..........................................
        gen_start_time = time.time()
        if self.fit_points_x is None:
            # latin hypercube sampling to get the initial sample of points
            points_x = latin_hypercube_sampling(
                n_sample=n_initial_points,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
            )
            # evaluate the target function
            points_y = target_function(points_x)

            self.fit_points_x = points_x
            self.fit_points_y = points_y
        else:
            points_x = self.fit_points_x
            points_y = self.fit_points_y

        self.step_gen_time.append(time.time() - gen_start_time)

        # fit the surrogate model
        start_time = time.time()

        for i in range(0, dim_out):
            s_i = self.surrogate_model()
            s_i.fit(points_x, points_y[:, i])
            self.surrogates.append(s_i)
            logger.info(
                f"Fitted the first surrogate model {i} in"
                f" {time.time() - start_time} sec."
            )

        # Additional class objects to help keep track of the sampling. `gen_time`
        # denotes the time to generate new points. `fit_time` is the time
        # to fit the surrogate.
        score = evaluate_modellist(
            self.logging_metrics,
            self.surrogates,
            self.test_points_x,
            self.test_points_y,
        )
        self.step_x.append(points_x)
        self.step_y.append(points_y)
        self.step_score.append(score[self.evaluation_metric.__name__])
        self.step_iter.append(0)
        self.step_fit_time.append(time.time() - start_time)

        # ..........................................
        # Iterative improvement (adaptive stage)
        # ..........................................
        for ii in range(max_n_iterations):
            logger.info(
                f"Started adaptive iteration step: {ii+1} (max steps:"
                f" {max_n_iterations})."
            )

            start_time = time.time()

            """
                The following implements the algorithm from [1] [2] [3]
            """

            # Step 1. Calculate the Voronoi cells
            random_points, distance_mx, closest_indicator_mx = calculate_voronoi_cells(
                points_x,
                domain_lower_bound=domain_lower_bound,
                domain_upper_bound=domain_upper_bound,
            )

            # Step 2. Determine the Voronoi cell with the highest error
            most_sensitive_cell = identify_sensitive_voronoi_cell(
                self.surrogates,
                self.surrogate_model,
                points_x,
                points_y,
                dim_out,
                self.test_points_x,
                self.test_points_y,
                5,
            )

            # Step 3. Pick the point within the most sensitive cell, furthest
            # away from point i.
            new_points_x = pick_new_samples(
                most_sensitive_cell,
                random_points,
                distance_mx,
                closest_indicator_mx,
                n=1,
            )

            self.step_gen_time.append(time.time() - start_time)
            logger.info(
                f"Found the next best {n_new_points_per_iteration} point(s) in "
                f"{time.time() - start_time} sec."
            )

            # evaluate the target function
            new_points_y = target_function(new_points_x)

            # add the new points to the old ones
            points_x = np.vstack((points_x, new_points_x))
            points_y = np.vstack((points_y, new_points_y.ravel()))

            # refit the surrogate
            start_time = time.time()
            for i, surrogate in enumerate(self.surrogates):
                # surrogate.update(new_points_x, new_points_y[i])
                surrogate.fit(points_x, points_y[:, i])
                self.step_fit_time.append(time.time() - start_time)
                logger.info(
                    f"Fitted a new surrogate model in {time.time() - start_time} sec."
                )

            self.fit_points_x = points_x
            self.fit_points_y = points_y

            score = evaluate_modellist(
                self.logging_metrics,
                self.surrogates,
                self.test_points_x,
                self.test_points_y,
            )
            logger.info(
                f"Fitted a new surrogate model in {time.time() - start_time} sec."
            )
            self.step_x.append(points_x)
            self.step_y.append(points_y)
            self.step_score.append(score[self.evaluation_metric.__name__])
            self.step_iter.append(ii + 1)
            timing_dict = {
                "Gen time": self.step_gen_time[ii + 1],
                "Fit time": self.step_fit_time[ii + 1],
            }
            write_scores(self.writer, score, ii + 1)
            write_timer(self.writer, timing_dict, ii + 1)

            self.score = score[self.evaluation_metric.__name__]
            self.iterations = ii
            # Save model every 200 iterations
            if self.iterations % 200 == 0:
                for s_i, s in enumerate(self.surrogates):
                    save_name = (
                        f"{self.run_name}_" f"out{s_i}_{self.iterations}_iters.pkl"
                    )
                    save_path = os.path.join(self.save_dir, save_name)
                    with open(save_path, "wb") as file:
                        pickle.dump(s, file)

            if self.score <= stopping_criterium:
                logger.info(f"Algorithm converged in {ii} iterations")
                # Save model if converged
                for s_i, s in enumerate(self.surrogates):
                    save_name = (
                        f"{self.run_name}_" f"out{s_i}_{self.iterations}_converged.pkl"
                    )
                    save_path = os.path.join(self.save_dir, save_name)
                    with open(save_path, "wb") as file:
                        pickle.dump(s, file)
                break
        self.writer.close()

        return self.fit_points_x, self.fit_points_y


def pick_new_samples(
    most_sensitive_cell, random_points, distance_mx, closest_indicator_mx, n=1
):
    sens_cell = closest_indicator_mx[:, most_sensitive_cell]
    voronoi_i_distances = distance_mx[sens_cell, most_sensitive_cell]
    voronoi_i_points = random_points[sens_cell, :]

    max_dist_i = np.argpartition(voronoi_i_distances, -n)[-1:]

    return voronoi_i_points[max_dist_i, :][:]


def identify_sensitive_voronoi_cell(
    surrogates: list,
    surrogate_model: object,
    points_X: np.ndarray,
    points_y: np.ndarray,
    n_dim_out: int,
    test_points_X: np.ndarray,
    test_points_y: np.ndarray,
    k=5,
):
    normalized_responses = []
    surrogates_nrmse = []

    for idx, m in enumerate(surrogates):
        normalized_responses.append(normalized_response(m, points_X))
        # equation 4 from [2]
        surrogates_nrmse.append(
            nrmse(m, test_points_X, test_points_y[:, idx])
        )  # equation 5
        # from [2]

    total_error = sum(surrogates_nrmse)
    weights = np.asarray([i / total_error for i in surrogates_nrmse])  # eq.
    # 15 from
    # [2]

    # nrmse_sys = max(surrogates_nrmse)  # equation 8 from [2]. not used.
    kfold = KFold(n_splits=k, random_state=None, shuffle=False)
    kfold_results = np.zeros((k, len(surrogates)))
    split_indices = []
    i = 0
    # CV approach from [3] to avoid costly surrogate building for higher
    # number of points
    for train_index, test_index in kfold.split(points_X):
        split_indices.append(train_index)
        for s in range(0, n_dim_out):
            X_train, X_test = points_X[train_index], points_X[test_index]
            y_train, y_test = points_y[train_index], points_y[test_index]

            s_i = surrogate_model()
            s_i.fit(X_train, y_train[:, s])
            y_pred = s_i.predict(X_test)
            kfold_results[i, s] = np.linalg.norm(y_test[:, s] - y_pred)
        kfold_results_multiout = np.sum(kfold_results, axis=1)

        i += 1

    worst_fold = np.argmax(kfold_results_multiout)  # identify the worst fold
    worst_point_indices = split_indices[worst_fold]
    worst_points_X = points_X[worst_point_indices]
    worst_points_y = points_y[worst_point_indices]
    cv_error_per_point = np.zeros((len(worst_points_X), n_dim_out))

    for i in range(len(worst_point_indices)):
        X_i = worst_points_X[i, :]
        y_i = worst_points_y[i]

        idx_exc_i = np.where(np.all(worst_points_X != X_i, axis=1))
        points_X_exc_i = worst_points_X[idx_exc_i]
        points_y_exc_i = worst_points_y[idx_exc_i]

        # implements #13 #14 from [2]
        for j in range(0, n_dim_out):
            s_i = surrogate_model()
            s_i.fit(points_X_exc_i, points_y_exc_i[:, j])
            # predict X[i] with surrogate
            y_pred = s_i.predict(X_i.reshape((1, -1)))

            cv_error_per_point[i, j] = np.linalg.norm(
                y_i[j] - y_pred
            )  # eq. 14 from [2]

    max_eij = np.max(cv_error_per_point)

    LOOCV_scores = np.sum(cv_error_per_point * weights + max_eij, axis=1)  #
    # eq. 16 from [2]

    return np.argmax(LOOCV_scores)


def calculate_voronoi_cells(
    points: np.ndarray,
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_simulation: int = None,
    random_points: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Estimate the Voronoi tessellation of `points` bounded by
    `domain_lower_bound` and `domain_upper_bound`.

    The algorithm is described in algorithm 1 of [1].

    Args:
        points:
            n_point x n_dim.
        domain_lower_bound:
            n_dim.
        domain_upper_bound:
            n_dim
        n_simulation:
            Number of random points used to estimate the relative volumes. If
            `random_points` is provided then this argument is ignored.
        random_points:

    Returns:
        Relative volumes in the same order as `points`.
    """
    # dimensions are not checked
    if n_simulation is None:
        n_simulation = 100 * points.shape[0] * points.shape[1]

    if random_points is None:
        n_dim = len(domain_lower_bound)
        random_points = domain_lower_bound + np.random.rand(n_simulation, n_dim) * (
            domain_upper_bound - domain_lower_bound
        )

    # all relevant distances, n_simulation x n_point
    distance_mx = cdist(random_points, points, metric="euclidean")

    # index of the closest `point` to each `random_point`
    col_idx_min = np.argmin(distance_mx, axis=1)
    row_idx_min = np.arange(distance_mx.shape[0])

    # indicator matrix to count the number of `random_points` closest to each `points`
    closest_indicator_mx = np.zeros(distance_mx.shape, dtype=bool)
    # place a one (True) in the indicator matrix if the element (distance) is a closest
    # distance
    closest_indicator_mx[row_idx_min, col_idx_min] = True

    return random_points, distance_mx, closest_indicator_mx



"""Test components of the implementation of the LOLA-Voronoi algorithm."""
import warnings

import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial import ConvexHull

from harlow.sampling.lola_voronoi import (
    best_neighborhoods,
    gradient_estimate,
    neighborhood_score,
    voronoi_volume_estimate,
)
from tests.unit_tests.bounded_voronoi_2d import bounded_voronoi_2d


# .......................................
# Best neighborhoods
# .......................................
def test_visually_2d_best_neighborhoods():
    """This test is to be checked by humans. Plot a few best neighborhoods in 2D."""
    # number of points to be used
    n_point = 10

    # Random points
    all_points_x = [np.random.rand(n_point, 2)]

    # Points along a circle perimeter and one in the center
    center_point_x = np.array([0, 0])
    radius = 2
    angles = np.random.rand(n_point - 1) * np.pi * 2
    perimeter_points_x = np.array(
        [
            center_point_x[0] + radius * np.cos(angles),
            center_point_x[1] + radius * np.sin(angles),
        ]
    ).T
    all_points_x.append(np.vstack((center_point_x, perimeter_points_x)))

    for points_x in all_points_x:
        # Find the best neighborhood for each row in `points_x`
        (
            best_neighborhood_scores,
            best_neighborhood_idxs,
            _,
            all_neighbor_point_idxs_combinations,
        ) = best_neighborhoods(points_x=points_x, n_point_last_iter=0)

        fig, axs = plt.subplots(2, 2, figsize=(10, 8))
        axs = axs.flatten()
        for ii, ax in enumerate(axs):
            # select reference point for plotting
            reference_point_x = points_x[ii]
            neighbor_points_x = points_x[
                all_neighbor_point_idxs_combinations[best_neighborhood_idxs[ii]], :
            ]
            ax.scatter(points_x[:, 0], points_x[:, 1])
            ax.scatter(
                reference_point_x[0], reference_point_x[1], label="reference point"
            )
            ax.plot(
                neighbor_points_x[:, 0],
                neighbor_points_x[:, 1],
                "s",
                markeredgecolor="red",
                markerfacecolor="None",
                label="best neighborhood points",
            )
            ax.set_xlabel("$x_1$")
            ax.set_ylabel("$x_2$")
            if ii == 0:
                ax.legend()
            ax.axis("equal")


# .......................................
# Neighborhood score
# .......................................
def test_example_of_section_4_2_3():
    # the 2D example from section 4.2.3 of [1]
    base_neighbor_points_x = np.array([[-1, 0], [1, 0], [0, -1]])
    reference_point_x = np.array([[0, 0]])

    expected_cpr_max_x = [0.0, 1.0]
    expected_ns_max_x = [0.0, 0.0]

    n_grid = 51
    x1_mx, x2_mx = np.meshgrid(np.linspace(-3, 3, n_grid), np.linspace(-3, 3, n_grid))
    added_neighbor_points_x = np.vstack((x1_mx.ravel(), x2_mx.ravel())).T

    ns_vec = np.empty(n_grid**2)
    cpr_vec = np.empty(n_grid**2)
    for ii, added_neighbor_point_x in enumerate(added_neighbor_points_x):
        neighbor_points_x = np.vstack((base_neighbor_points_x, added_neighbor_point_x))
        ns, cpr = neighborhood_score(
            neighbor_points_x=neighbor_points_x, reference_point_x=reference_point_x
        )
        ns_vec[ii] = ns
        cpr_vec[ii] = cpr

    ns_mx = ns_vec.reshape((n_grid, n_grid))
    cpr_mx = cpr_vec.reshape((n_grid, n_grid))

    # check the maximum of cross-polytope ratio
    idx = np.unravel_index(np.argmax(cpr_mx), cpr_mx.shape)
    cpr_max_x = [x1_mx[idx], x2_mx[idx]]
    a_tol = 0.5 * 6 / (n_grid - 1)
    np.testing.assert_allclose(expected_cpr_max_x, cpr_max_x, atol=a_tol)

    # check the maximum of cross-polytope ratio
    idx = np.unravel_index(np.argmax(cpr_mx), cpr_mx.shape)
    cpr_max_x = [x1_mx[idx], x2_mx[idx]]
    a_tol = 0.5 * 6 / (n_grid - 1)
    np.testing.assert_allclose(expected_cpr_max_x, cpr_max_x, atol=a_tol)
    assert np.all(np.logical_and(0.0 <= cpr_mx, cpr_mx <= 1.0))

    idx = np.unravel_index(np.argmax(ns_mx), ns_mx.shape)
    ns_max_x = [x1_mx[idx], x2_mx[idx]]
    np.testing.assert_allclose(expected_ns_max_x, ns_max_x, atol=a_tol)

    # ..........................
    # Plot (for visual check)
    # ..........................
    # Compare with Fig(4.5) of [1]
    cmap = "viridis"
    to_plot = {
        "Cross-polytope ratio": {"values_mx": cpr_mx, "maximum_point_x": cpr_max_x},
        "Neighborhood score": {"values_mx": ns_mx, "maximum_point_x": ns_max_x},
    }
    fig, axs = plt.subplots(2, 2, figsize=(10, 8))

    for ii, (data_key, data_value) in enumerate(to_plot.items()):
        ax_left = axs[ii, 0]
        ax_right = axs[ii, 1]
        z_mx = data_value["values_mx"]
        max_x = data_value["maximum_point_x"]

        # left, surface plot
        ax_left.remove()
        ax_left = fig.add_subplot(2, 2, 2 * ii + 1, projection="3d")
        ax_left.azim = -160

        ax_left.plot_surface(x1_mx, x2_mx, z_mx, rstride=1, cstride=1, cmap=cmap)
        ax_left.set_xlabel("$x_1$")
        ax_left.set_ylabel("$x_2$")
        ax_left.set_title(data_key)

        # right, contour plot
        ax_right.contourf(x1_mx, x2_mx, z_mx, cmap=cmap)
        ax_right.scatter(
            base_neighbor_points_x[:, 0],
            base_neighbor_points_x[:, 1],
            label="neighbor points",
            marker="s",
        )
        ax_right.scatter(
            reference_point_x[:, 0], reference_point_x[:, 1], label="reference point"
        )
        ax_right.scatter(max_x[0], max_x[1], marker="d", label="maximum")
        ax_right.set_xlabel("$x_1$")
        ax_right.set_ylabel("$x_2$")
        ax_right.set_title(data_key)
        ax_right.legend()


def test_circle_cross_polytope_ratio():
    # perfect cross-polytope
    radius = 2
    n_neighbor = 2 * 2
    reference_points_x = [np.array([0, 0]), np.array([3, 1])]
    cpr_expected = 1.0

    for reference_point_x in reference_points_x:
        ii = np.arange(n_neighbor)
        x1 = reference_point_x[0] + radius * np.cos(np.pi / n_neighbor * (1 + 2 * ii))
        x2 = reference_point_x[1] + radius * np.sin(np.pi / n_neighbor * (1 + 2 * ii))
        neighbor_points_x = np.vstack((x1, x2)).T

        _, cpr = neighborhood_score(
            neighbor_points_x=neighbor_points_x, reference_point_x=reference_point_x
        )

        np.testing.assert_almost_equal(cpr_expected, cpr)


def test_imperfect_polytope_ratio():
    n_dims = [1, 2, 5, 10, 15]
    cpr_max = 1.0
    cpr_min = 0.0

    for n_dim in n_dims:
        n_neighbor = n_dim * 2
        reference_point_x = np.ones((1, n_dim))

        neighbor_points_x = np.random.rand(n_neighbor, n_dim) + reference_point_x

        _, cpr = neighborhood_score(
            neighbor_points_x=neighbor_points_x, reference_point_x=reference_point_x
        )

        assert cpr_min <= cpr <= cpr_max


def test_worst_polytope_ratio():
    # all neighbors in one point
    n_dims = [1, 2, 5, 10, 15]
    cpr_expected = 0.0

    for n_dim in n_dims:
        n_neighbor = n_dim * 2
        reference_point_x = np.ones((1, n_dim))

        neighbor_points_x = (
            np.tile(np.random.rand(1, n_dim), (n_neighbor, 1)) + reference_point_x
        )

        _, cpr = neighborhood_score(
            neighbor_points_x=neighbor_points_x, reference_point_x=reference_point_x
        )

        np.testing.assert_almost_equal(cpr_expected, cpr)


# .......................................
# Voronoi volume estimation
# .......................................
def test_voronoi_volume_2d():
    """Test the Monte Carlo simulation based general code against an exact 2D
    solution."""
    domain_lower_bound = np.array([-0.1, -1])
    domain_upper_bound = np.array([1.1, 2])
    n_simulation = int(1e6)
    plot_fig = False

    n_points = [2, 3, 5, 10, 20]
    for n_point in n_points:
        points = domain_lower_bound + np.random.rand(n_point, 2) * (
            domain_upper_bound - domain_lower_bound
        )

        # Reference solution
        vor = bounded_voronoi_2d(
            points,
            bounding_box=[
                domain_lower_bound[0],
                domain_upper_bound[0],
                domain_lower_bound[1],
                domain_upper_bound[1],
            ],
        )

        areas_expected = []
        for filtered_region in vor.filtered_regions:
            areas_expected.append(ConvexHull(vor.vertices[filtered_region, :]).volume)

        domain_area = (domain_upper_bound[0] - domain_lower_bound[0]) * (
            domain_upper_bound[1] - domain_lower_bound[1]
        )

        # The reference algorithm sometimes fails to cover the entire domain,
        # we skip the check if that happens.
        if not np.isclose(sum(areas_expected), domain_area, atol=1e-4):
            warnings.warn(
                f"n_point={n_point}."
                "\nThe reference algorithm has failed to cover the "
                "entire domain. This check is skipped."
            )
            continue

        if plot_fig:
            fig, ax = plt.subplots()
            # Plot initial points
            ax.plot(vor.filtered_points[:, 0], vor.filtered_points[:, 1], "b.")
            # Plot ridges points
            for region in vor.filtered_regions:
                vertices = vor.vertices[region, :]
                ax.plot(vertices[:, 0], vertices[:, 1], "go")
            # Plot ridges
            for region in vor.filtered_regions:
                vertices = vor.vertices[region + [region[0]], :]
                ax.plot(vertices[:, 0], vertices[:, 1], "k-")

        relative_areas_expected = np.array(areas_expected) / np.sum(areas_expected)

        # Tested implementation (general, for N-dimensions)
        relative_areas = voronoi_volume_estimate(
            points=points,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
            n_simulation=n_simulation,
        )[0]

        # until we can get the order right..
        # np.testing.assert_array_almost_equal(
        #     relative_areas_expected, relative_areas, decimal=3
        # )
        np.testing.assert_array_almost_equal(
            np.sort(relative_areas_expected), np.sort(relative_areas), decimal=3
        )


# .......................................
# Gradient estimation
# .......................................
def test_gradient_linear():
    """If the points are from a hyperplane then the gradient estimate should be
    exact."""
    n_dims = [1, 2, 5, 10]

    for n_dim in n_dims:
        grad_expected = np.arange(1, n_dim + 1)

        def hyperplane(x, grad):
            return np.sum(grad * x, axis=1, keepdims=True)

        reference_point_x = np.random.randn(1, n_dim)
        reference_point_y = float(hyperplane(reference_point_x, grad_expected))

        neighbor_points_x = np.random.randn(2 * n_dim, n_dim)
        neighbor_points_y = hyperplane(neighbor_points_x, grad_expected)

        grad = gradient_estimate(
            reference_point_x=reference_point_x,
            reference_point_y=reference_point_y,
            neighbor_points_x=neighbor_points_x,
            neighbor_points_y=neighbor_points_y,
        ).ravel()

        np.testing.assert_array_almost_equal(grad_expected, grad)



import time

import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial import distance

from harlow.utils.distance import pdist_condensed, pdist_full_matrix


def test_pdist_condensed():
    n_points = [10, 100, 1000]
    n_repeats = [100, 10, 5]
    n_dim = 4

    plot_fig = True
    compare_runtime = False

    t_harlow = np.empty(len(n_points))
    t_scipy = np.empty(len(n_points))
    for ii, (n_point, n_repeat) in enumerate(zip(n_points, n_repeats)):
        x_mx = np.random.rand(n_point, n_dim)

        dist = pdist_condensed(x_mx)
        dist_expected = distance.pdist(x_mx)

        # check if the correct values are returned
        np.testing.assert_array_almost_equal(dist, dist_expected)

        # runtime measurement
        if compare_runtime:
            t0 = time.time()
            for _ in range(n_repeat):
                pdist_condensed(x_mx)
            t_harlow[ii] = (time.time() - t0) / n_repeat

            t0 = time.time()
            for _ in range(n_repeat):
                distance.pdist(x_mx)
            t_scipy[ii] = (time.time() - t0) / n_repeat

    if compare_runtime and plot_fig:
        fig, ax = plt.subplots()
        ax.plot(n_points, t_scipy / t_harlow, "o-")
        ax.set_xscale("log")
        ax.set_xlabel("Number of vectors (points)")
        ax.set_ylabel("$t_\\mathrm{scipy}/t_\\mathrm{harlow}$")
        ax.set_title("`pdist` runtime with condensed output")
        ax.grid()


def test_pdist_full_matrix():
    n_points = [10, 100, 1000]
    n_repeats = [100, 10, 5]
    n_dim = 4

    plot_fig = True
    compare_runtime = False

    t_harlow = np.empty(len(n_points))
    t_scipy = np.empty(len(n_points))
    for ii, (n_point, n_repeat) in enumerate(zip(n_points, n_repeats)):
        x_mx = np.random.rand(n_point, n_dim)

        dist = pdist_full_matrix(x_mx)
        dist_expected = distance.squareform(distance.pdist(x_mx))

        # check if the correct values are returned
        np.testing.assert_array_almost_equal(dist, dist_expected)

        # runtime measurement
        if compare_runtime:
            t0 = time.time()
            for _ in range(n_repeat):
                pdist_full_matrix(x_mx)
            t_harlow[ii] = (time.time() - t0) / n_repeat

            t0 = time.time()
            for _ in range(n_repeat):
                distance.squareform(distance.pdist(x_mx))
            t_scipy[ii] = (time.time() - t0) / n_repeat

    if compare_runtime and plot_fig:
        fig, ax = plt.subplots()
        ax.plot(n_points, t_scipy / t_harlow, "o-")
        ax.set_xscale("log")
        ax.set_xlabel("Number of vectors (points)")
        ax.set_ylabel("$t_\\mathrm{scipy}/t_\\mathrm{harlow}$")
        ax.set_title("`pdist` runtime with full mx output")
        ax.grid()



"""Test whether the samples match (or close to) the Matlab SUMO samples.
To reduce runtime a dummy surrogate model class is used: `EmptySurrogate`."""

import numpy as np

from harlow.sampling import LolaVoronoi
from harlow.utils.helper_functions import latin_hypercube_sampling
from harlow.utils.test_functions import forrester_1d, lldeh_1d
from tests.integration_tests.utils import plot_1d_lola_voronoi


class EmptySurrogate:
    def __init__(self):
        pass

    def update(self, x, y):
        pass

    def fit(self, x, y):
        return self.update(x, y)

    @staticmethod
    def predict(x):
        return np.zeros(x.shape[0]).reshape((-1, 1))


def test_forrester_1d_against_sumo():
    n_new_point_per_iteration = 1
    n_iter = 15

    start_points_x = np.array([[0.0, 0.25, 0.50, 0.75, 1.0]]).T
    domain_lower_bound = np.array([0])
    domain_upper_bound = np.array([1])
    plot_fig = True

    def target_function(x: np.ndarray):
        return forrester_1d(x).ravel()

    test_X = latin_hypercube_sampling(domain_lower_bound, domain_upper_bound, 100)
    test_y = target_function(test_X).reshape((-1, 1))

    start_points_y = target_function(start_points_x).reshape((-1, 1))
    surrogate_model = EmptySurrogate()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=target_function,
        surrogate_model=surrogate_model,
        fit_points_x=start_points_x,
        fit_points_y=start_points_y,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        test_points_x=test_X,
        test_points_y=test_y,
    )
    lv.sample(
        max_n_iterations=n_iter,
        n_new_points_per_iteration=n_new_point_per_iteration,
    )

    print(np.sort(lv.fit_points_x[len(start_points_x) :, :], axis=0))
    # ............................
    # Check accuracy
    # ............................
    # TODO: compare with SUMO results

    # ............................
    # Plot
    # ............................
    if plot_fig:
        plot_1d_lola_voronoi(
            lv,
            n_initial_point=len(start_points_x),
            n_new_point_per_iteration=n_new_point_per_iteration,
        )


def test_lldeh_1d_against_sumo():
    n_new_point_per_iteration = 1
    n_iter = 15

    domain_lower_bound = np.array([-20])
    domain_upper_bound = np.array([20])
    start_points_x = np.linspace(domain_lower_bound, domain_upper_bound, 5)

    plot_fig = True

    def target_function(x: np.ndarray):
        return lldeh_1d(x, a=2.1)

    test_X = latin_hypercube_sampling(
        domain_lower_bound, domain_upper_bound, 100
    ).reshape((-1, 1))
    test_y = target_function(test_X).reshape((-1, 1))

    start_points_y = target_function(start_points_x)
    surrogate_model = EmptySurrogate()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=target_function,
        surrogate_model=surrogate_model,
        fit_points_x=start_points_x,
        fit_points_y=start_points_y,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        test_points_x=test_X,
        test_points_y=test_y,
    )
    lv.sample(
        max_n_iterations=n_iter,
        n_new_points_per_iteration=n_new_point_per_iteration,
    )

    print(np.sort(lv.fit_points_x[len(start_points_x) :, :], axis=0))
    # ............................
    # Check accuracy
    # ............................
    # TODO: compare with SUMO results

    # ............................
    # Plot
    # ............................
    if plot_fig:
        _, ax = plot_1d_lola_voronoi(
            lv,
            n_initial_point=len(start_points_x),
            n_new_point_per_iteration=n_new_point_per_iteration,
            n_grid_point=500,
        )
        ax.set_aspect("equal", "box")
        ax.legend(bbox_to_anchor=(0.5, 1.05))


if __name__ == "__main__":
    test_lldeh_1d_against_sumo()
    test_forrester_1d_against_sumo()



"""Benchmarktest from Crombecq et al. (2011). Benchmarks LV and
Probabilistic sampling vs random sampling. """
import json
import math
import time

import numpy as np
from loguru import logger
from sklearn.metrics import mean_squared_error

from harlow.sampling.lola_voronoi import LolaVoronoi
from harlow.sampling.probabilistic_sampling import Probabilistic_sampler
from harlow.surrogating.surrogate_model import VanillaGaussianProcess
from harlow.utils.helper_functions import latin_hypercube_sampling
from harlow.utils.test_functions import peaks_2d

domains_lower_bound = [np.array([-3, -3]), np.array([-5, -5]), np.array([-8, -8])]
domains_upper_bound = [np.array([3, 3]), np.array([5, 5]), np.array([8, 8])]
n_initial_point = 10
n_new_points_per_iteration = 1
rmse_criterium = 0.05


def rmse(x, y):
    return math.sqrt(mean_squared_error(x, y))


def create_test_set(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = peaks_2d(test_X).reshape((-1, 1))

    return test_X, test_y


def run_benchmark():
    lv_results = []
    ps_results = []
    random_results = []

    for i in range(0, len(domains_upper_bound)):
        test_X, test_y = create_test_set(
            domains_lower_bound[i], domains_upper_bound[i], 500
        )

        start_points_X = latin_hypercube_sampling(
            domains_lower_bound[i], domains_upper_bound[i], n_initial_point
        )
        start_points_y = peaks_2d(start_points_X).reshape((-1, 1))

        lv_results.append(
            test_LV_sampling(
                start_points_X,
                start_points_y,
                domains_lower_bound[i],
                domains_upper_bound[i],
                test_X,
                test_y,
            )
        )

        ps_results.append(
            test_probabilistic_sampling(
                start_points_X,
                start_points_y,
                domains_lower_bound[i],
                domains_upper_bound[i],
                test_X,
                test_y,
            )
        )

        random_results.append(
            test_random_sampling(
                start_points_X,
                start_points_y,
                domains_lower_bound[i],
                domains_upper_bound[i],
                test_X,
                test_y,
            )
        )

    with open("LV_results", "w") as fout:
        json.dump(lv_results, fout)

    with open("PS_results", "w") as fout:
        json.dump(ps_results, fout)

    with open("RAND_results", "w") as fout:
        json.dump(random_results, fout)


def test_LV_sampling(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
):

    surrogate_model = VanillaGaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=peaks_2d,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        evaluation_metric=rmse,
    )
    main_start = time.time()
    lv.sample(
        max_n_iterations=None,
        n_initial_points=n_initial_point,
        n_new_points_per_iteration=n_new_points_per_iteration,
        stopping_criterium=rmse_criterium,
    )

    return {
        "domain_lower_bound": domain_lower_bound.tolist(),
        "domain_upper_bound": domain_upper_bound.tolist(),
        "iterations": lv.iterations,
        "score": lv.score,
        "elapsed_time": time.time() - main_start,
    }


def test_probabilistic_sampling(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
):

    surrogate_model = VanillaGaussianProcess()

    # ............................
    # Surrogating
    # ............................
    ps = Probabilistic_sampler(
        target_function=peaks_2d,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        evaluation_metric=rmse,
    )
    main_start = time.time()
    ps.sample(
        n_iter=None,
        n_initial_point=n_initial_point,
        stopping_criterium=rmse_criterium,
    )

    return {
        "domain_lower_bound": domain_lower_bound.tolist(),
        "domain_upper_bound": domain_upper_bound.tolist(),
        "iterations": ps.iterations,
        "score": ps.score,
        "elapsed_time": time.time() - main_start,
    }


def test_random_sampling(
    points_x, points_y, domain_lower_bound, domain_upper_bound, test_X, test_y
):

    main_start = time.time()

    start_time = time.time()
    surrogate_model = VanillaGaussianProcess()
    surrogate_model.fit(points_x, points_y)
    logger.info(f"Fitted a new surrogate model in {time.time() - start_time} sec.")

    score = rmse(surrogate_model.predict(test_X), test_y)
    iteration = 1
    while score > rmse_criterium:

        X_new = latin_hypercube_sampling(
            n_sample=1,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
        )
        y_new = peaks_2d(X_new).reshape((-1, 1))
        points_x = np.concatenate((points_x, X_new))
        points_y = np.concatenate((points_y, y_new))

        start_time = time.time()
        surrogate_model.fit(points_x, points_y)
        logger.info(f"Fitted a surrogate model in {time.time() - start_time} sec.")
        iteration += 1
        score = rmse(surrogate_model.predict(test_X), test_y)
        logger.info(f"Score {score}")

    logger.info(f"Algorithm converged in {iteration} iterations")
    logger.info(f"Algorithm converged with score {score}")

    return {
        "domain_lower_bound": domain_lower_bound.tolist(),
        "domain_upper_bound": domain_upper_bound.tolist(),
        "iterations": iteration,
        "score": score,
        "elapsed_time": time.time() - main_start,
    }


if __name__ == "__main__":
    run_benchmark()



from harlow.surrogating.surrogate_model import (  # noqa F401; GaussianProcessRegression,
    BatchIndependentGaussianProcess,
    BayesianNeuralNetwork,
    DeepKernelMultiTaskGaussianProcess,
    GaussianProcessTFP,
    ModelListGaussianProcess,
    MultiTaskGaussianProcess,
    NeuralNetwork,
    Surrogate,
    VanillaGaussianProcess,
)



def write_scores(writer, score_dict, iteration):
    # Writer log & various metric scores
    for k, v in score_dict.items():
        for i in range(len(v)):
            name = f"{k} surrogate {i}"
            writer.add_scalar(name, v[i], iteration)


def write_timer(writer, time_dict, iteration):
    for k, v in time_dict.items():
        writer.add_scalar(k, v, iteration)



"""
Comparison of Lola-Voronoi and Fuzzy Lola-Voronoi for different dimensions.
The different methods are compared for a given random seed to determine:

* The surrogating wall clock time.
* The loss of efficiency caused by the heuristic, if any.
* That the optimizations are implemented correctly.
"""

import argparse
import json
import os

import numpy as np

from harlow.sampling import (
    CVVoronoi,
    FuzzyLolaVoronoi,
    LatinHypercube,
    LolaVoronoi,
    ProbabilisticSampler,
)
from harlow.surrogating.surrogate_model import VanillaGaussianProcess
from harlow.utils.helper_functions import latin_hypercube_sampling
from harlow.utils.metrics import mae, rmse, rrse
from harlow.utils.test_functions import (
    hartmann,
    peaks_2d,
    peaks_2d_multivariate,
    stybtang,
)

np.random.seed(0)
stop_thresh = 0.01  # For RMSE or 0.005 - 0.0025


def create_test_set_2D(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = peaks_2d(test_X).reshape((-1, 1))

    return test_X, test_y


def create_test_set_2Dmulti(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = peaks_2d_multivariate(test_X)

    return test_X, test_y


def create_test_set_6D(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = hartmann(test_X).reshape((-1, 1))

    return test_X, test_y


def create_test_set_8D(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = stybtang(test_X).reshape((-1, 1))

    return test_X, test_y


def dict_to_array(list_of_dicts, key):
    arr_0 = list_of_dicts[0][key][-1]
    for _, lst in enumerate(list_of_dicts[1::]):
        lst = lst[key]
        arr_0 = np.vstack((arr_0, lst[-1]))
    return arr_0


def steps_to_array(list_of_dicts, key):
    arr_0 = list_of_dicts[0][key]
    for _, lst in enumerate(list_of_dicts[1::]):
        lst = lst[key]
        arr_0 = np.vstack((arr_0, lst))
    return arr_0


def run_benchmark(
    name,
    evaluation_metric,
    logging_metrics,
    method,
    adapt_steps,
    n_new_points,
    problem,
    n_initial_point,
    test_size,
):
    sampling_res_list = []
    save_path = os.path.join("saves", name)
    os.makedirs(save_path, exist_ok=True)

    if problem == 2:
        domains_lower_bound = np.array([-8.0, -8.0])
        domains_upper_bound = np.array([8.0, 8.0])
        test_X, test_y = create_test_set_2D(
            domains_lower_bound, domains_upper_bound, test_size
        )
        start_points_X = latin_hypercube_sampling(
            domains_lower_bound, domains_upper_bound, n_initial_point
        )
        start_points_y = peaks_2d(start_points_X).reshape((-1, 1))
        target_func = peaks_2d

    elif problem == 6:
        domains_lower_bound = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        domains_upper_bound = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
        test_X, test_y = create_test_set_6D(
            domains_lower_bound, domains_upper_bound, test_size
        )

        start_points_X = latin_hypercube_sampling(
            domains_lower_bound, domains_upper_bound, n_initial_point
        )
        start_points_y = hartmann(start_points_X).reshape((-1, 1))
        target_func = hartmann

    elif problem == 8:
        domains_lower_bound = np.array([-5.0, -5.0, -5.0, -5.0, -5.0, -5.0, -5.0, -5.0])
        domains_upper_bound = np.array([5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0])
        test_X, test_y = create_test_set_8D(
            domains_lower_bound, domains_upper_bound, test_size
        )
        # print('TEST_X, TEST_y', test_X, test_X.shape, test_y, test_y.shape)
        start_points_X = latin_hypercube_sampling(
            domains_lower_bound, domains_upper_bound, n_initial_point
        )
        start_points_y = stybtang(start_points_X).reshape((-1, 1))
        # print('START_X, START_y', start_points_X, start_points_X.shape,
        # start_points_y, start_points_y.shape)
        target_func = stybtang
    elif problem == -2:
        domains_lower_bound = np.array([-8.0, -8.0])
        domains_upper_bound = np.array([8.0, 8.0])
        test_X, test_y = create_test_set_2Dmulti(
            domains_lower_bound, domains_upper_bound, test_size
        )
        start_points_X = latin_hypercube_sampling(
            domains_lower_bound, domains_upper_bound, n_initial_point
        )
        start_points_y = peaks_2d_multivariate(start_points_X)

        target_func = peaks_2d_multivariate

    sampling_run_res = test_sampler(
        start_points_X,
        start_points_y,
        domains_lower_bound,
        domains_upper_bound,
        test_X,
        test_y,
        evaluation_metric,
        logging_metrics,
        method,
        adapt_steps,
        n_initial_point,
        n_new_points,
        target_func,
        name,
        save_path,
    )

    sampling_res_list.append(sampling_run_res)

    # Save to json for plotting etc !!
    json_name = "{}_results_with_score_{}".format(
        name, sampling_run_res.get("score")[-1]
    )
    json_save_path = os.path.join("json_saves", json_name)
    os.makedirs("json_saves", exist_ok=True)
    with open(f"{json_save_path}.json", "w") as fout:
        sampling_run_res["step_x"] = [i.tolist() for i in sampling_run_res["step_x"]]
        sampling_run_res["step_y"] = [i.tolist() for i in sampling_run_res["step_y"]]
        sampling_run_res["score"] = [i for i in sampling_run_res["score"]]
        json.dump(sampling_run_res, fout)

    return sampling_res_list


def test_sampler(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    evaluation_metric,
    logging_metrics,
    sampler,
    n_iter_sampling,
    n_initial_point,
    n_new_points_per_iteration,
    target_f,
    name,
    save_path,
):

    surrogate_model = VanillaGaussianProcess
    # ............................
    # Surrogating
    # ............................
    if sampler == "FLOLA":
        lv = FuzzyLolaVoronoi(
            target_function=target_f,
            surrogate_model=surrogate_model,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
            fit_points_x=start_points_X,
            fit_points_y=start_points_y,
            test_points_x=test_X,
            test_points_y=test_y,
            evaluation_metric=evaluation_metric,
            logging_metrics=logging_metrics,
            run_name=name,
            save_dir=save_path,
        )
    elif sampler == "LOLA":
        lv = LolaVoronoi(
            target_function=target_f,
            surrogate_model=surrogate_model,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
            fit_points_x=start_points_X,
            fit_points_y=start_points_y,
            test_points_x=test_X,
            test_points_y=test_y,
            evaluation_metric=evaluation_metric,
            logging_metrics=logging_metrics,
            run_name=name,
            save_dir=save_path,
        )
    elif sampler == "Prob":
        lv = ProbabilisticSampler(
            target_function=target_f,
            surrogate_model=surrogate_model,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
            fit_points_x=start_points_X,
            fit_points_y=start_points_y,
            test_points_x=test_X,
            test_points_y=test_y,
            evaluation_metric=evaluation_metric,
            logging_metrics=logging_metrics,
            run_name=name,
        )
    elif sampler == "Random":
        lv = LatinHypercube(
            target_function=target_f,
            surrogate_model=surrogate_model,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
            fit_points_x=start_points_X,
            fit_points_y=start_points_y,
            test_points_x=test_X,
            test_points_y=test_y,
            evaluation_metric=evaluation_metric,
            logging_metrics=logging_metrics,
            run_name=name,
        )

    elif sampler == "CV":
        lv = CVVoronoi(
            target_function=target_f,
            surrogate_model=surrogate_model,
            domain_lower_bound=domain_lower_bound,
            domain_upper_bound=domain_upper_bound,
            fit_points_x=start_points_X,
            fit_points_y=start_points_y,
            test_points_x=test_X,
            test_points_y=test_y,
            evaluation_metric=evaluation_metric,
            logging_metrics=logging_metrics,
            run_name=name,
            save_dir=save_path,
        )
    lv.sample(
        n_initial_points=n_initial_point,
        max_n_iterations=n_iter_sampling,
        n_new_points_per_iteration=n_new_points_per_iteration,
        stopping_criterium=stop_thresh,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s", "--sampler", default="FLOLA", type=str, help="Name of the sampler"
    )
    parser.add_argument(
        "-p",
        "--problem",
        default=6,
        type=int,
        help="Dimensionality of the problem to solve",
    )
    parser.add_argument(
        "-i",
        "--init_p",
        default=15,
        type=int,
        help="Number of initial points to start sampling",
    )
    parser.add_argument(
        "-em",
        "--evaluation_metric",
        default="rmse",
        type=str,
        help="Define the evaluation metric function",
    )
    parser.add_argument(
        "-lm",
        "--logging_metrics",
        nargs="+",
        help="Define the logging metrics functions",
    )
    parser.add_argument(
        "-st",
        "--steps",
        default=3000,
        type=int,
        help="Number of iterative adaptive sampling steps",
    )
    parser.add_argument(
        "-n",
        "--n_points_iter",
        default=1,
        type=int,
        help="Number of points we add at every adaptive sampling steps",
    )
    args = parser.parse_args()

    TEST_SIZE = 500
    evaluation_metric = rmse
    if args.logging_metrics == "all":
        logging_metrics = [rrse, mae]
    else:
        logging_metrics = []

    run_name = "Bench_{}_with_{}_initial_points_on_{}D_problem".format(
        args.sampler, args.init_p, args.problem
    )
    print(run_name)

    flv_sampling_out = run_benchmark(
        run_name,
        evaluation_metric,
        logging_metrics,
        args.sampler,
        args.steps,
        args.n_points_iter,
        args.problem,
        args.init_p,
        TEST_SIZE,
    )



"""
Adaptive sampling based on uncertainty heuristics.
This algorithm is from:
    Xuzheng Chai (2019) Probabilistic system identification and reliability updating
    for hydraulic structures - Application to sheet pile walls

Adapted from implementation in Prob_Taralli:
    https://gitlab.com/tno-bim/taralli/-/blob/d82a5f42e918f4864d4d6f18f0dbdf8c1f2799c6/
    prob_taralli/surrogating/adaptive_infill_gpr.py

"""
import time
from typing import Callable

import numpy as np
from loguru import logger
from scipy.optimize import differential_evolution

from harlow.sampling.sampling_baseclass import Sampler
from harlow.surrogating.surrogate_model import Surrogate
from harlow.utils.helper_functions import evaluate, latin_hypercube_sampling
from harlow.utils.log_writer import write_scores, write_timer
from harlow.utils.metrics import rmse


class ProbabilisticSampler(Sampler):
    def __init__(
        self,
        target_function: Callable[[np.ndarray], np.ndarray],
        surrogate_model: Surrogate,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = rmse,
        logging_metrics: list = None,
        verbose: bool = False,
        run_name: str = None,
        save_dir: str = "",
    ):

        super(ProbabilisticSampler, self).__init__(
            target_function,
            surrogate_model,
            domain_lower_bound,
            domain_upper_bound,
            fit_points_x,
            fit_points_y,
            test_points_x,
            test_points_y,
            evaluation_metric,
            logging_metrics,
            verbose,
            run_name,
            save_dir,
        )

        self.iterations = 0

    def sample(
        self,
        n_initial_points: int = 20,
        n_new_points_per_iteration: int = 1,
        stopping_criterium: float = None,
        max_n_iterations: int = 5000,
        epsilon: float = 0.005,
    ):

        n_dim = len(self.domain_lower_bound)

        if n_initial_points is None:
            n_initial_point = 5 * n_dim

        logger.info(f"Adaptive sampling iteration {self.iterations}.")

        if not self.surrogate_model.is_probabilistic:
            raise NotImplementedError(
                "Uncertainty based sampling only implemented for probabilistic \
                surrogate models."
            )

        # ..........................................
        # Initial sample of points
        # ..........................................
        gen_start_time = time.time()
        if self.fit_points_x is None:
            points_x = latin_hypercube_sampling(
                n_sample=n_initial_point,
                domain_lower_bound=self.domain_lower_bound,
                domain_upper_bound=self.domain_upper_bound,
            )
            points_y = self.target_function(points_x)
        else:
            points_x = self.fit_points_x
            points_y = self.fit_points_y
        self.step_gen_time.append(time.time() - gen_start_time)

        start_time = time.time()
        self.surrogate_model.fit(points_x, points_y)
        logger.info(f"Fitted a new surrogate model in {time.time() - start_time} sec.")
        self.step_fit_time.append(time.time() - gen_start_time)

        convergence = False

        while convergence is False:
            bounds = [
                (self.domain_lower_bound[i], self.domain_upper_bound[i])
                for i in range(n_dim)
            ]

            def prediction_std(x):
                if x.ndim == 1:
                    x = np.expand_dims(x, axis=0)

                std = -(
                    self.surrogate_model.predict(x, return_std=True)[1]
                    # - self.surrogate_model.noise_std
                )

                return std

            start_time = time.time()
            diff_evolution_result = differential_evolution(
                prediction_std, bounds=bounds
            )
            logger.info(
                f"Finished differential evolution in {time.time() - start_time} sec."
            )
            std_max = -diff_evolution_result.fun
            self.step_gen_time.append(time.time() - start_time)

            x_new = np.expand_dims(diff_evolution_result.x, axis=0)
            y_new = self.target_function(x_new)

            start_time = time.time()
            self.surrogate_model.update(x_new, y_new)
            logger.info(
                f"Fitted a new surrogate model in {time.time() - start_time} sec."
            )
            self.step_fit_time.append(time.time() - gen_start_time)

            points_x = np.concatenate((points_x, x_new))
            points_y = np.concatenate((points_y, y_new))

            if std_max <= epsilon or self.iterations > max_n_iterations:

                logger.info("std_max <= epsilon or max iterations reached")
                convergence = True
            elif stopping_criterium:
                predicted_y = self.surrogate_model.predict(
                    self.test_points_x, as_array=True
                )
                score = evaluate(self.logging_metrics, self.test_points_y, predicted_y)
                self.score = score[self.evaluation_metric.__name__]
                logger.info(f"Evaluation metric score on provided testset: {score}")
                if self.score <= stopping_criterium:
                    self.number_of_iterations_at_convergence = self.iterations
                    logger.info(f"Algorithm converged in {self.iterations} iterations")
                    convergence = True

            self.fit_points_x = points_x
            self.fit_points_y = points_y
            self.step_x.append(points_x)
            self.step_y.append(points_y)
            self.step_score.append(self.score)
            self.step_iter.append(self.iterations + 1)

            timing_dict = {
                "Gen time": self.step_gen_time[self.iterations + 1],
                "Fit time": self.step_fit_time[self.iterations + 1],
            }
            write_scores(self.writer, score, self.iterations + 1)
            write_timer(self.writer, timing_dict, self.iterations + 1)

            self.score = score
            self.iterations += 1

        self.writer.close()

        return self.fit_points_x, self.fit_points_y



"""Fast and numba compatible scipy.distance like functions.
Inspired by and based on: https://github.com/talboger/fastdist"""

import numpy as np
from numba import jit, prange

nopython = True
fastmath = True
parallel = True
cache = True


@jit(nopython=nopython, fastmath=fastmath, cache=cache)
def euclidean_distance(u_vec: np.ndarray, v_vec: np.ndarray) -> float:
    """Euclidean distance between two vectors."""
    dist = 0
    for ii in range(len(u_vec)):
        dist += (u_vec[ii] - v_vec[ii]) ** 2
    return np.sqrt(dist)


def fractional_distance(u_vec: np.ndarray, v_vec: np.ndarray) -> float:
    """Fractional distance between two vectors."""
    dist = 0
    ndim = 2
    for ii in range(len(u_vec)):
        dist += (u_vec[ii] - v_vec[ii]) ** ndim
    return np.pow(1.0 / ndim)


@jit(nopython=nopython, fastmath=fastmath, parallel=parallel, cache=cache)
def pdist_condensed(x_mx: np.ndarray) -> np.ndarray:
    """
    Pairwise Euclidean distances between n-dimensional vectors. Similar to
    `scipy.spatial.distance.pdist` but faster for a large number of vectors.

    TODO: maybe it could be faster if there was just one loop, kk -> ii, jj.

    Args:
        x_mx:
            An `m` by `n` matrix of `m` vectors in an `n`-dimensional space.

    Returns:
        Condensed distance matrix.
    """
    n_point = x_mx.shape[0]
    n_elem = int((n_point**2 - n_point) / 2)
    dist = np.empty(n_elem)
    for ii in prange(n_point):
        for jj in prange(ii + 1, n_point):
            kk = int(ii * n_point - (ii**2 + 3 * ii) / 2 + jj - 1)
            dist[kk] = euclidean_distance(x_mx[ii], x_mx[jj])
    return dist


@jit(nopython=nopython, fastmath=fastmath, parallel=parallel, cache=cache)
def pdist_full_matrix(x_mx: np.ndarray) -> np.ndarray:
    """
    Pairwise Euclidean distances between n-dimensional vectors. Similar to
    `scipy.spatial.distance.pdist` and `scipy.spatial.distance.squareform` but faster
    for a large number of vectors.

    Args:
        x_mx:
            An `m` by `n` matrix of `m` vectors in an `n`-dimensional space.

    Returns:
        A full (`m` by `m`) distance matrix.
    """
    n_point = x_mx.shape[0]
    dist = np.zeros((n_point, n_point))
    for ii in prange(n_point):
        for jj in prange(ii + 1, n_point):
            dist[ii, jj] = euclidean_distance(x_mx[ii], x_mx[jj])
    # make it symmetric
    return dist + dist.T


@jit(nopython=nopython, fastmath=fastmath, parallel=parallel, cache=cache)
def fractional_pdist_full_matrix(x_mx: np.ndarray) -> np.ndarray:
    """
    Pairwise Euclidean distances between n-dimensional vectors. Similar to
    `scipy.spatial.distance.pdist` and `scipy.spatial.distance.squareform` but faster
    for a large number of vectors.

    Args:
        x_mx:
            An `m` by `n` matrix of `m` vectors in an `n`-dimensional space.

    Returns:
        A full (`m` by `m`) distance matrix.
    """

    n_point = x_mx.shape[0]
    dist = np.zeros((n_point, n_point))
    for ii in prange(n_point):
        for jj in prange(ii + 1, n_point):
            dist[ii, jj] = fractional_distance(x_mx[ii], x_mx[jj])
    # make it symmetric
    return dist + dist.T



"""
Forward and reverse transforms typically used in surrogating and optimization.

Notes:
    * It may be usefull to define a convention for in/output shapes to ensure that
    the different surrogating, sampling, transformation and vizualization tools adhere
    to a common structure for the input and output data.
    * For now we assume that features `X` have size (n_batch x n_points x n_features)
    and outputs `y` have size (n_batch x n_points x n_outputs).

TODO:
    * More thorough testing
    * Add docstrings
    * Add type hints
"""
from abc import ABC, abstractmethod

import numpy as np
import torch


class Transform(ABC):
    """
    Abstract base class for the input and output transforms. Implements the forward
    and reverse transform methods.
    """

    @abstractmethod
    def forward(self, X):
        raise NotImplementedError

    @abstractmethod
    def reverse(self, X):
        raise NotImplementedError


class ChainTransform(Transform):
    """
    Class used to chain together a series of transforms

    TODO:
        * Should this make a copy of `X` first or change it inplace?
    """

    def __init__(self, *args):
        self.lst_transforms = list(args)

    def forward(self, X):
        for transf in self.lst_transforms:
            X = transf.forward(X)
        return X

    def reverse(self, X):
        for transf in self.lst_transforms.__reversed__():
            X = transf.reverse(X)
        return X


class Identity(Transform):
    """
    Defines the identity transform that can be used as a placeholder
    if no transforms are specified.
    """

    def __init__(self):
        pass

    def forward(self, X):
        return X

    def reverse(self, X):
        return X


class TensorTransform(Transform):
    """
    Transforms an input `numpy.ndarray` to `torch.tensor` and vice-versa
    """

    def __init__(self, target_type=torch.float32):
        self.target_type = target_type

    def forward(self, X):
        return torch.from_numpy(X).type(self.target_type)

    def reverse(self, X):
        return X.detach().cpu().numpy().astype(np.float64)


class Standardize(Transform):
    """
    Scales features to zero mean and unit standard deviation.
    """

    def __init__(self):
        self.mean = None
        self.std = None

    def forward(self, X):
        self.mean = np.expand_dims(X.mean(axis=-2), axis=-2)
        self.std = np.expand_dims(X.std(axis=-2), axis=-2)

        return (X - self.mean) / self.std

    def reverse(self, X, mean=None, std=None):

        if mean is not None:
            self.mean = mean

        if std is not None:
            self.std = std

        if (self.mean is None) or (self.std is None):
            raise RuntimeError(
                "The mean or standard deviation are `None`. These"
                "must either be initialized by calling the forward"
                "transform or provided as keyword arguments."
            )

        return self.mean + self.std * X


class Normalize(Transform):
    """
    Applies min-max scaling to given features.
    """

    def __init__(self):
        self.min = None
        self.max = None

    def forward(self, X):
        self.min = np.expand_dims(X.min(axis=-2), axis=-2)
        self.max = np.expand_dims(X.max(axis=-2), axis=-2)

        return (X - self.min) / (self.max - self.min)

    def reverse(self, X, xmin=None, xmax=None):

        if xmin is not None:
            self.min = xmin

        if xmax is not None:
            self.max = xmax

        if (self.min is None) or (self.max is None):
            raise RuntimeError(
                "The min or max are `None`. These"
                "must either be initialized by calling the forward"
                "transform or provided as keyword arguments."
            )

        return X * (self.max - self.min) + self.min


class ExpandDims(Transform):
    """
    Expands a two-dimensional feature array `X` to three dimensions,
    or expands a one-dimensional target array `y` to two dimensions.
    """

    def __init__(self):
        pass

    def forward(self, X):
        input_dims = X.ndim

        # If input is one-dimensional, assume it is a vector of target
        # values
        if input_dims == 1:
            return np.expand_dims(X, 0).T

        # If input is two-dimensional, assume it is an array of input
        # values
        elif input_dims == 2:
            return np.expand_dims(X, 0)

        # Otherwise raise error
        else:
            raise ValueError(
                f"Input array must be one or two-dimensional "
                f"but has {input_dims} dimensions."
            )

    def reverse(self, X):
        return X.squeeze()



"""
Helper functions for the adaptive sampling strategies.
"""

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from skopt.sampler import Lhs
from skopt.space import Space

tfd = tfp.distributions


def evaluate_modellist(metric, model, test_points_X, test_points_y):
    """
    Evaluate user specified metric for the current iteration

    Returns:
    """
    count_model = 0
    metric_dict = {}
    if not isinstance(metric, list):
        metric = [metric]
    if metric is None or test_points_X is None:
        raise ValueError
    else:
        for metric_fun in metric:
            scores = []
            for m in model:
                scores.append(
                    metric_fun(m.predict(test_points_X), test_points_y[:, count_model])
                )
            metric_dict[metric_fun.__name__] = scores

    return metric_dict


def evaluate(metric, true_y, predicted_y):
    """
    Evaluate user specified metric for the current iteration

    Returns:
    """
    metric_dict = {}
    if not isinstance(metric, list):
        metric = [metric]
    if metric is None or true_y is None:
        for metric_fun in metric:
            metric_dict[metric_fun.__name__] = []
    else:
        for metric_fun in metric:
            scores = []
            for m in range(predicted_y.shape[1]):
                scores.append(metric_fun(true_y[:, m], predicted_y[:, m]))
            metric_dict[metric_fun.__name__] = scores

    return metric_dict


def normalized_response(model: object, X: np.ndarray):
    preds = model.predict(X)
    return (model.predict(X) - np.min(preds)) / (np.max(preds) - np.min(preds))


def NLL(y, distr):
    return -distr.log_prob(y)


def normal_sp(params):
    return tfd.Normal(
        loc=params[:, 0:1], scale=1e-3 + tf.math.softplus(0.05 * params[:, 1:2])
    )  # both parameters are learnable


def latin_hypercube_sampling(
    domain_lower_bound: np.ndarray,
    domain_upper_bound: np.ndarray,
    n_sample: int,
    method="maximin",
):
    domain = np.vstack((domain_lower_bound, domain_upper_bound)).astype(float).T
    space = Space(list(map(tuple, domain)))
    lhs = Lhs(criterion=method, iterations=5000)
    samples = lhs.generate(space.dimensions, n_sample)

    return np.array(samples, dtype=float)



"""
Comparison of different methods for reducing the optimzal neighbourhood
calculation time. The different methods are compared for a given random
seed to determine:

* The surrogating wall clock time.
* The loss of efficiency caused by the heuristic, if any.
* That the optimizations are implemented correctly.
"""

import math

import numpy as np
from sklearn.metrics import mean_squared_error

from harlow.sampling import FuzzyLolaVoronoi
from harlow.surrogating import VanillaGaussianProcess
from harlow.utils.helper_functions import latin_hypercube_sampling
from harlow.utils.test_functions import peaks_2d

domains_lower_bound = np.array([-8, -8])
domains_upper_bound = np.array([8, 8])
n_initial_point = 15
n_new_points_per_iteration = 1
rmse_criterium = 0.001
np.random.seed(123)
n_iter_sampling = 30
n_iter_runs = 100


def rmse(x, y):
    return math.sqrt(mean_squared_error(x, y))


def create_test_set(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = peaks_2d(test_X).reshape((-1, 1))

    return test_X, test_y


def run_2D_test():
    fuzzy_sampling_results = []
    test_X, test_y = create_test_set(domains_lower_bound, domains_upper_bound, 500)

    start_points_X = latin_hypercube_sampling(
        domains_lower_bound, domains_upper_bound, n_initial_point
    )
    start_points_y = peaks_2d(start_points_X).reshape((-1, 1))

    fuzzy_sampling_results.append(
        test_2D_fuzzy_sampling(
            start_points_X,
            start_points_y,
            domains_lower_bound,
            domains_upper_bound,
            test_X,
            test_y,
            n_iter_sampling,
        )
    )


def test_2D_fuzzy_sampling(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    n_iter,
):

    surrogate_model = VanillaGaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = FuzzyLolaVoronoi(
        target_function=peaks_2d,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        evaluation_metric=rmse,
    )
    # main_start = time.time()
    lv.sample(
        n_iter=n_iter,
        n_initial_point=n_initial_point,
        n_new_point_per_iteration=n_new_points_per_iteration,
        ignore_old_neighborhoods=False,
        ignore_far_neighborhoods=False,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


if __name__ == "__main__":
    run_2D_test()



"""
Comparison of different methods for reducing the optimzal neighbourhood
calculation time. The different methods are compared for a given random
seed to determine:

* The surrogating wall clock time.
* The loss of efficiency caused by the heuristic, if any.
* That the optimizations are implemented correctly.
"""

import math

import numpy as np
from matplotlib import pyplot as plt
from sklearn.metrics import mean_squared_error

from harlow.helper_functions import latin_hypercube_sampling
from harlow.lola_voronoi import LolaVoronoi
from harlow.surrogate_model import GaussianProcess
from harlow.utils.test_functions import peaks_2d

domains_lower_bound = np.array([-8, -8])
domains_upper_bound = np.array([8, 8])
n_initial_point = 10
n_new_points_per_iteration = 1
rmse_criterium = 0.001
np.random.seed(123)
n_iter_sampling = 25
n_iter_runs = 100


def rmse(x, y):
    return math.sqrt(mean_squared_error(x, y))


def create_test_set(min_domain, max_domain, n):
    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = peaks_2d(test_X).reshape((-1, 1))

    return test_X, test_y


def run_benchmark():
    full_sampling_results = []
    median_sampling_results = []
    new_sampling_results = []
    both_sampling_results = []

    for _i in range(n_iter_runs):
        test_X, test_y = create_test_set(domains_lower_bound, domains_upper_bound, 500)

        start_points_X = latin_hypercube_sampling(
            domains_lower_bound, domains_upper_bound, n_initial_point
        )
        start_points_y = peaks_2d(start_points_X).reshape((-1, 1))

        full_sampling_results.append(
            test_sampling_full(
                start_points_X,
                start_points_y,
                domains_lower_bound,
                domains_upper_bound,
                test_X,
                test_y,
                n_iter_sampling,
            )
        )

        median_sampling_results.append(
            test_sampling_median(
                start_points_X,
                start_points_y,
                domains_lower_bound,
                domains_upper_bound,
                test_X,
                test_y,
                n_iter_sampling,
            )
        )

        new_sampling_results.append(
            test_sampling_new(
                start_points_X,
                start_points_y,
                domains_lower_bound,
                domains_upper_bound,
                test_X,
                test_y,
                n_iter_sampling,
            )
        )

        both_sampling_results.append(
            test_sampling_both(
                start_points_X,
                start_points_y,
                domains_lower_bound,
                domains_upper_bound,
                test_X,
                test_y,
                n_iter_sampling,
            )
        )

    return (
        full_sampling_results,
        median_sampling_results,
        new_sampling_results,
        both_sampling_results,
    )


def test_sampling_full(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    n_iter,
):

    surrogate_model = GaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=peaks_2d,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        evaluation_metric=rmse,
    )
    # main_start = time.time()
    lv.sample(
        n_iter=n_iter,
        n_initial_point=n_initial_point,
        n_new_point_per_iteration=n_new_points_per_iteration,
        ignore_old_neighborhoods=False,
        ignore_far_neighborhoods=False,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


def test_sampling_median(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    n_iter,
):

    surrogate_model = GaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=peaks_2d,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        evaluation_metric=rmse,
    )
    # main_start = time.time()
    lv.sample(
        n_iter=n_iter,
        n_initial_point=n_initial_point,
        n_new_point_per_iteration=n_new_points_per_iteration,
        ignore_old_neighborhoods=False,
        ignore_far_neighborhoods=True,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


def test_sampling_new(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    n_iter,
):

    surrogate_model = GaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=peaks_2d,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        evaluation_metric=rmse,
    )
    # main_start = time.time()
    lv.sample(
        n_iter=n_iter,
        n_initial_point=n_initial_point,
        n_new_point_per_iteration=n_new_points_per_iteration,
        ignore_old_neighborhoods=True,
        ignore_far_neighborhoods=False,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


def test_sampling_both(
    start_points_X,
    start_points_y,
    domain_lower_bound,
    domain_upper_bound,
    test_X,
    test_y,
    n_iter,
):

    surrogate_model = GaussianProcess()

    # ............................
    # Surrogating
    # ............................
    lv = LolaVoronoi(
        target_function=peaks_2d,
        surrogate_model=surrogate_model,
        domain_lower_bound=domain_lower_bound,
        domain_upper_bound=domain_upper_bound,
        fit_points_x=start_points_X,
        fit_points_y=start_points_y,
        test_points_x=test_X,
        test_points_y=test_y,
        evaluation_metric=rmse,
    )
    # main_start = time.time()
    lv.sample(
        n_iter=n_iter,
        n_initial_point=n_initial_point,
        n_new_point_per_iteration=n_new_points_per_iteration,
        ignore_old_neighborhoods=True,
        ignore_far_neighborhoods=True,
    )

    return {
        "step_x": lv.step_x,
        "step_y": lv.step_y,
        "score": lv.step_score,
        "iteration": lv.step_iter,
        "gen_time": lv.step_gen_time,
        "fit_time": lv.step_fit_time,
    }


(
    full_sampling_out,
    median_sampling_out,
    new_sampling_out,
    both_sampling_out,
) = run_benchmark()


# =======================================================================
# PLOTTING
# =======================================================================

# Meshgrid of function solution
npts_grid = 100
x_min = -8.0
x_max = 8.0
x = np.linspace(x_min, x_max, npts_grid)
y = np.linspace(x_min, x_max, npts_grid)
Z = np.zeros((npts_grid, npts_grid))
full_samples_list = []
median_sampling_list = []
new_sampling_list = []
both_sampling_list = []

for idx_x, xi in enumerate(x):
    for idx_y, yi in enumerate(y):
        Z[idx_x, idx_y] = peaks_2d([xi, yi])


def dict_to_array(list_of_dicts, key):
    arr_0 = list_of_dicts[0][key][-1]
    for _, lst in enumerate(list_of_dicts[1::]):
        lst = lst[key]
        arr_0 = np.vstack((arr_0, lst[-1]))
    return arr_0


def steps_to_array(list_of_dicts, key):
    arr_0 = list_of_dicts[0][key]
    for _, lst in enumerate(list_of_dicts[1::]):
        lst = lst[key]
        arr_0 = np.vstack((arr_0, lst))
    return arr_0


iterations = steps_to_array(full_sampling_out, "iteration")
ql = 0.1
qh = 1.0 - ql
fig, ax = plt.subplots(3, 1, figsize=(12, 9))

full_score = steps_to_array(full_sampling_out, "score")
median_score = steps_to_array(median_sampling_out, "score")
new_score = steps_to_array(new_sampling_out, "score")
both_score = steps_to_array(both_sampling_out, "score")
full_time = steps_to_array(full_sampling_out, "gen_time")
median_time = steps_to_array(median_sampling_out, "gen_time")
new_time = steps_to_array(new_sampling_out, "gen_time")
both_time = steps_to_array(both_sampling_out, "gen_time")

# Mean score
ax[0].plot(np.mean(full_score, axis=0), label="Full")
ax[0].plot(np.mean(median_score, axis=0), label="Median", linestyle="dashed")
ax[0].plot(np.mean(new_score, axis=0), label="New", linestyle="dashed")
ax[0].plot(np.mean(both_score, axis=0), label="Both", linestyle="dashed")
ax[0].set_ylabel("RMSE")

# Score quantiles
ax[0].fill_between(
    iterations[0],
    np.quantile(full_score, ql, axis=0),
    np.quantile(full_score, qh, axis=0),
    alpha=0.5,
)
ax[0].fill_between(
    iterations[0],
    np.quantile(median_score, ql, axis=0),
    np.quantile(median_score, qh, axis=0),
    linestyle="dashed",
    alpha=0.5,
)
ax[0].fill_between(
    iterations[0],
    np.quantile(new_score, ql, axis=0),
    np.quantile(new_score, qh, axis=0),
    linestyle="dashed",
    alpha=0.5,
)
ax[0].fill_between(
    iterations[0],
    np.quantile(both_score, ql, axis=0),
    np.quantile(both_score, qh, axis=0),
    linestyle="dashed",
    alpha=0.5,
)
ax[0].set_ylabel("RMSE")

# Mean generation time
ax[1].plot(np.mean(full_time, axis=0), label="Full")
ax[1].plot(np.mean(median_time, axis=0), label="Median")
ax[1].plot(np.mean(new_time, axis=0), label="New")
ax[1].plot(np.mean(both_time, axis=0), label="Both")
ax[1].set_xlabel("Iteration no.")
ax[1].set_ylabel("Gen. time [s]")

# Generation time quantiles
ax[1].fill_between(
    iterations[0],
    np.quantile(full_time, ql, axis=0),
    np.quantile(full_time, qh, axis=0),
    alpha=0.5,
)
ax[1].fill_between(
    iterations[0],
    np.quantile(median_time, ql, axis=0),
    np.quantile(median_time, qh, axis=0),
    alpha=0.5,
)
ax[1].fill_between(
    iterations[0],
    np.quantile(new_time, ql, axis=0),
    np.quantile(new_time, qh, axis=0),
    alpha=0.5,
)
ax[1].fill_between(
    iterations[0],
    np.quantile(both_time, ql, axis=0),
    np.quantile(both_time, qh, axis=0),
    alpha=0.5,
)
ax[1].set_xlabel("Iteration no.")
ax[1].set_ylabel("Gen. time [s]")

# Test points scatter plot for all iterations
ax[2].contourf(x, y, Z)
ax[2].scatter(
    dict_to_array(full_sampling_out, "step_x")[:, 0],
    dict_to_array(full_sampling_out, "step_x")[:, 1],
    label="Full",
    alpha=0.5,
)
ax[2].scatter(
    dict_to_array(median_sampling_out, "step_x")[:, 0],
    dict_to_array(median_sampling_out, "step_x")[:, 1],
    label="Median",
    alpha=0.5,
)
ax[2].scatter(
    dict_to_array(new_sampling_out, "step_x")[:, 0],
    dict_to_array(new_sampling_out, "step_x")[:, 1],
    label="New",
    alpha=0.5,
)
ax[2].scatter(
    dict_to_array(both_sampling_out, "step_x")[:, 0],
    dict_to_array(both_sampling_out, "step_x")[:, 1],
    label="Both",
    alpha=0.5,
)


ax[0].legend()
ax[0].grid()
ax[1].legend()
ax[1].grid()
ax[2].legend()
plt.show()



"""
Comparison of different Multi-task GPR approaches:

* ModelList GPs: A list of independent Gaussian process surrogates
* BatchIndependent GPs: For independent outputs using the same covariance and likelihood
* MultiTask GPs: For learning similarities between outcomes
"""

from timeit import default_timer as timer

import numpy as np
import torch

# from botorch.models.transforms import Normalize, Standardize
from matplotlib import pyplot as plt
from sklearn.metrics import mean_squared_error

from harlow.surrogating import (
    BatchIndependentGaussianProcess,
    DeepKernelMultiTaskGaussianProcess,
    ModelListGaussianProcess,
    MultiTaskGaussianProcess,
)
from harlow.utils.examples.model_twin_girder_betti import (  # noqa: I201
    IJssel_bridge_model,
)
from harlow.utils.helper_functions import latin_hypercube_sampling

# ====================================================================
# HELPER FUNCTIONS
# ====================================================================


def create_test_set(min_domain, max_domain, n):

    test_X = latin_hypercube_sampling(min_domain, max_domain, n)
    test_y = response(test_X, sensor_positions)

    return torch.FloatTensor(test_X), torch.FloatTensor(test_y)


def rmse(x, y):
    return mean_squared_error(x, y, squared=False)


def get_param_idx(params_dict):
    return {key: idx_key for idx_key, key in enumerate(params_dict)}


# ====================================================================
# SURROGATING PARAMETERS
# ====================================================================
N_train = 200
N_update = 100
N_test = 100
N_pred = 100
N_iter = 1000
rmse_criterium = 0.1
silence_warnings = True

# ====================================================================
# INITIALIZE MODEL
# ====================================================================
# Set model parameters
meas_type = "fugro"
E = 210e6
max_elem_length = 2.0 * 1e3

# Sensor names and positions
sensor_names = ["H1_S", "H2_S", "H3_S", "H4_S", "H5_S", "H7_S", "H8_S", "H9_S", "H10_S"]
sensor_positions = [
    20.42,
    34.82,
    47.700,
    61.970,
    68.600,
    96.800,
    113.9,
    123.900,
    147.500,
]

# All parameters. This is the order that parameters will be expcted in
# within the various functions in this scrípt (e.g. model response function)
params_all = ["Kr1", "Kr2", "Kr3", "Kr4", "Kv"]

N_tasks = len(sensor_positions)
N_features = len(params_all)

# Parameters that are shared between models
params_common = [
    "Kv",
]

# Parameters that are not shared by all models
params_model = {
    "H1_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H2_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H3_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H4_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H5_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H7_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H8_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H9_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
    "H10_S": ["Kr1", "Kr2", "Kr3", "Kr4"],
}

# Define FE models and append to list
models_dict = {
    idx_model: IJssel_bridge_model(
        sname, E, max_elem_length=max_elem_length, truck_load=meas_type
    )
    for idx_model, sname in enumerate(sensor_names)
}
node_xs = models_dict[0].node_xs

# # ====================================================================
# # DOMAIN BOUNDS
# # ====================================================================

# Prior for first support rotational stiffness Kr1
Kr_low = 4.0
Kr_high = 10.0

# Ground truth and prior for K-brace spring vertical stiffness
Kv_low = 0.0
Kv_high = 8.0

c = -0.1754

# Bounds of the time domain
t_low = np.min(node_xs)
t_high = np.max(node_xs)

params_priors = {
    "Kv": {"dist": "uniform", "low": Kv_low, "high": Kv_high},
    "Kr1": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
    "Kr2": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
    "Kr3": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
    "Kr4": {"dist": "uniform", "low": Kr_low, "high": Kr_high},
}

# Create domain bounds
domain_lower_bound = np.array([params_priors[param]["low"] for param in params_all])
domain_upper_bound = np.array([params_priors[param]["high"] for param in params_all])

# Bounds as tensor
bounds = torch.tensor(np.vstack([domain_lower_bound, domain_upper_bound]))

# Indices of all params
param_idx = get_param_idx(params_all)

# Build list of parameter indices per model
list_params = [
    [param_idx[param] for param in params_model[key] + params_common]
    for key in params_model.keys()
]


# # ====================================================================
# # MODEL FUNCTION
# # ====================================================================
def response(X, pts):

    # Initialize
    X = np.atleast_2d(X)
    N_x = X.shape[0]
    N_y = len(pts)
    res = np.zeros((N_x, N_y))

    print(f"Evaluating response function at {N_x} points for {N_y} outputs")

    # Iterate over parameter array
    for idx_x, x in enumerate(X):
        Kr1 = x[param_idx["Kr1"]]
        Kr2 = x[param_idx["Kr2"]]
        Kr3 = x[param_idx["Kr3"]]
        Kr4 = x[param_idx["Kr4"]]
        Kv = x[param_idx["Kv"]]

        # Rotational stiffness input array
        arr_Kr = np.repeat(np.array([Kr1, Kr2, Kr3, Kr4]), 2)
        arr_Kr = np.append(arr_Kr, np.zeros(4))

        # Iterate over models and evaluate
        for idx_t, model_key in enumerate(models_dict.keys()):

            # Get model and prediction t
            t = pts[idx_t]
            model = models_dict[model_key]

            # Evaluate each model in list
            res[idx_x, idx_t] = np.interp(
                t,
                model.node_xs,
                model.il_stress_truckload(c, lane="left", Kr=10**arr_Kr, Kv=10**Kv),
            )

    return res


# Model function taking only X as input, to be passed to the sampler.
def func_model(X):
    return response(X, sensor_positions)


# ====================================================================
# STANDARDIZING AND NORMALIZING
# ====================================================================
def input_transform(X):
    return X  # Normalize(d=N_features, bounds=bounds)(X)


def output_transform(y):
    return y  # Standardize(m=N_tasks)(y)[0]


# ====================================================================
# GENERATE TEST AND TRAIN DATA
# ====================================================================
# Each column of train_Y corresponds to one GP
print(f"Create training set N = {N_train}:")
train_X, train_y = create_test_set(domain_lower_bound, domain_upper_bound, N_train)
# test_X, test_y = create_test_set(domain_lower_bound, domain_upper_bound, N_test)

# print(f"Create update set N = {N_update}:")
# # To check surrogate updating
# update_X, update_y = create_test_set(domain_lower_bound, domain_upper_bound, N_update)

# Check if machine is GPU compatible and assign the device to use
device = torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")
# device = torch.device("cpu")
print("Training on Device: {}".format(device))

# gpu=False
# ====================================================================
# DEFINE SURROGATE
# ====================================================================
N_tasks = train_y.shape[1]

surrogate_MLGP = ModelListGaussianProcess(
    input_transform(train_X),
    output_transform(train_y),
    model_names=sensor_names,
    list_params=list_params,
    training_max_iter=N_iter,
    silence_warnings=silence_warnings,
    dev=device,
)

surrogate_BIGP = BatchIndependentGaussianProcess(
    input_transform(train_X),
    output_transform(train_y),
    num_tasks=N_tasks,
    training_max_iter=N_iter,
    silence_warnings=silence_warnings,
    dev=device,
)

surrogate_MTGP = MultiTaskGaussianProcess(
    input_transform(train_X),
    output_transform(train_y),
    num_tasks=N_tasks,
    training_max_iter=N_iter,
    silence_warnings=silence_warnings,
    dev=device,
)

surrogate_DKLGP = DeepKernelMultiTaskGaussianProcess(
    input_transform(train_X),
    output_transform(train_y),
    num_tasks=N_tasks,
    training_max_iter=N_iter,
    silence_warnings=silence_warnings,
    dev=device,
)


# Create a list of surrogates
list_surrogates = [surrogate_MLGP, surrogate_BIGP, surrogate_MTGP, surrogate_DKLGP]
list_GP_type = [
    "ModelList GP",
    "BatchIndependent GP",
    "MultiTask GP",
    "DeepKernelMultiTask GP",
]

# ====================================================================
# FIT
# ====================================================================
print("============= Initial fit =====================")
for i, surrogate_i in enumerate(list_surrogates):
    t1 = timer()
    print("Fitting: " + list_GP_type[i])
    surrogate_i.fit(train_X, train_y)
    t2 = timer()
    print("Fitted " + list_GP_type[i] + f" in {t2 - t1} sec.")
    print("--------------------------------------------------")

# ====================================================================
# UPDATE
# ====================================================================
# print("============= Update =====================")
# for i, surrogate_i in enumerate(list_surrogates):
#     print("Updating: " + list_GP_type[i])
#     surrogate_i.update(update_X, update_y)

# ====================================================================
# SURROGATE PREDICT
# ====================================================================

# Tensor of prediction points
vec_Kv = np.linspace(Kv_low, Kv_high, N_pred)
pred_X = np.tile(np.array([7.0, 7.0, 7.0, 7.0]), (N_pred, 1))
pred_X = np.hstack((pred_X, vec_Kv.reshape(-1, 1)))
pred_X = torch.FloatTensor(pred_X)

# Physical model prediction
print(f"Create test set: N = {N_pred}")
true_y = response(pred_X, sensor_positions)

print("============= Predict =====================")
# Surrogate model prediction
y_pred = []
for i, surrogate_i in enumerate(list_surrogates):
    print("Predicting: " + list_GP_type[i])
    y_pred.append(surrogate_i.predict(input_transform(pred_X), return_std=False))
    print("--------------------------------------------------")

# Initialize plots
nrows = 3
ncols = int(np.ceil(N_tasks / 3))

for i, surrogate_i in enumerate(list_surrogates):
    print("Plotting: " + list_GP_type[i])
    print("--------------------------------------------------")

    # Initialize plot
    f, axes = plt.subplots(nrows, ncols, figsize=(3 * ncols, 3 * nrows))

    # Get surrogate output
    mean_i = surrogate_i.mean
    upper_i = surrogate_i.cr_u
    lower_i = surrogate_i.cr_l
    train_X_i = surrogate_i.train_X.cpu().numpy()
    train_y_i = surrogate_i.train_y.cpu().numpy()
    for j, ax_i in enumerate(axes.ravel()):
        grid_idx = np.unravel_index(j, (nrows, ncols))

        # Plot training data as black stars
        ax_i.plot(
            input_transform(train_X)[:, -1],
            output_transform(train_y)[:, j],
            "k*",
            label="Observations",
        )

        # Predictive mean as blue line
        ax_i.plot(
            input_transform(pred_X)[:, -1].numpy(),
            output_transform(mean_i)[:, j].cpu().numpy(),
            "b",
            label="Mean",
        )

        # Shade in confidence
        ax_i.fill_between(
            input_transform(pred_X)[:, -1].numpy(),
            output_transform(lower_i)[:, j].cpu().detach().numpy(),
            output_transform(upper_i)[:, j].cpu().detach().numpy(),
            alpha=0.5,
            label="Confidence",
        )
        ax_i.plot(
            input_transform(pred_X)[:, -1].numpy(),
            output_transform(torch.from_numpy(true_y))[:, j],
            color="red",
            linestyle="dashed",
            label="Model",
        )
        ax_i.set_title(f"Sensor: {sensor_names[j]}")

    axes[0, 0].legend()
    plt.suptitle("Model: " + list_GP_type[i])
    plt.show()


plt.figure()
for i, surrogate_i in enumerate(list_surrogates):
    plt.plot(surrogate_i.vec_loss, label=list_GP_type[i])
    plt.xlabel("N_iter")
    plt.ylabel("Loss")
plt.legend()
plt.grid()
plt.show()



import math

import numpy as np
from sklearn.metrics import mean_absolute_error, mean_squared_error


# TODO RRSE code should be checked for correctness
def _rrse(x, y):
    dims = x.shape
    x_bar = np.mean(y)
    x_bar_arr = np.zeros(dims)
    x_bar_arr.fill(x_bar)

    return math.sqrt(mean_squared_error(x, y) / mean_squared_error(x, x_bar_arr))


def nrmse(model: object, test_points_X: np.ndarray, test_points_y: np.ndarray):
    """normalized root mean square error"""
    preds = model.predict(test_points_X)

    return np.sqrt(
        1
        / len(test_points_y)
        * np.sum(
            np.square(
                (test_points_y - preds)
                / (np.max(test_points_y) - np.min(test_points_y))
            )
        )
    )


def rrse(actual: np.ndarray, predicted: np.ndarray):
    """Root Relative Squared Error"""
    return np.sqrt(
        np.sum(np.square(actual - predicted))
        / np.sum(np.square(actual - np.mean(actual)))
    )


def rmse(x, y):
    return math.sqrt(mean_squared_error(x, y))


def mae(x, y):
    return mean_absolute_error(x, y)



"""
Active sampling test using BoTorch and the Predictive Variance acquisition
function.

The aim is to:

1. Use active learning to efficiently sample and surrogate an expensive
to evaluate model.
    * Relevant to the development of efficient surrogating methods in
    ERP DT WP.A.
    * Based on: https://botorch.org/tutorials/closed_loop_botorch_only

2. Extend to the case where we have a ModelListGP surrogate, i.e. a surrogate
of a multi-output model with each output being an independent task.
    * This is one of the surrogate models that would be interesting to
    test in the Moerdijk case of ERP DT WP.B.

Usefull:
    * Implementation in ax:
    https://github.com/facebook/Ax/issues/460

    * Issues with qNegIntegratedPosteriorVariance:
    https://github.com/pytorch/botorch/issues/573
"""

import os
import time
from typing import Callable

import numpy as np
import torch
from botorch.acquisition.active_learning import qNegIntegratedPosteriorVariance
from botorch.optim import optimize_acqf
from loguru import logger

from harlow.sampling.sampling_baseclass import Sampler
from harlow.surrogating.surrogate_model import Surrogate
from harlow.utils.helper_functions import latin_hypercube_sampling
from harlow.utils.metrics import rmse

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
dtype = torch.double
SMOKE_TEST = os.environ.get("SMOKE_TEST")


class NegativeIntegratedPosteriorVarianceSampler(Sampler):
    """
    NOTE: This sampler is experimental and currently only works with GPyTorch surrogates
    """

    def __init__(
        self,
        target_function: Callable[[np.ndarray], np.ndarray],
        surrogate_model: Surrogate,
        domain_lower_bound: np.ndarray,
        domain_upper_bound: np.ndarray,
        fit_points_x: np.ndarray = None,
        fit_points_y: np.ndarray = None,
        test_points_x: np.ndarray = None,
        test_points_y: np.ndarray = None,
        evaluation_metric: Callable = rmse,
        logging_metrics: list = None,
        verbose: bool = False,
        run_name: str = None,
        save_dir: str = "",
        n_mc_points: int = 256,
        q: int = 1,
        num_restarts: int = 10,
        raw_samples: int = 512,
        optimizer_options: dict = None,
        func_sample: Callable = latin_hypercube_sampling,
    ):

        super(NegativeIntegratedPosteriorVarianceSampler, self).__init__(
            target_function,
            surrogate_model,
            domain_lower_bound,
            domain_upper_bound,
            fit_points_x,
            fit_points_y,
            test_points_x,
            test_points_y,
            evaluation_metric,
            logging_metrics,
            verbose,
            run_name,
            save_dir,
        )

        # Sampler specific initialization
        self.n_mc_points = n_mc_points
        self.q = q
        self.num_restarts = num_restarts
        self.raw_samples = raw_samples
        self.func_sample = func_sample
        self.metric = evaluation_metric
        self.verbose = verbose
        self.mc_samples = None

        # Search space
        self.bounds = torch.tensor(np.vstack((domain_lower_bound, domain_upper_bound)))
        self.ndim = self.bounds.shape[-1]

        # Check if sampling function is provided, otherwise use default
        self.func_sample = lambda n: func_sample(
            n_sample=n,
            domain_lower_bound=self.domain_lower_bound,
            domain_upper_bound=self.domain_upper_bound,
        )

        # Optimizer kwargs
        self.optimizer_options = optimizer_options
        if not self.optimizer_options:
            self.optimizer_options = {"batch_limit": 5, "maxiter": 200}

        # Check surrogate model
        if not self.surrogate_model.is_probabilistic:
            raise NotImplementedError(
                "Uncertainty based sampling only implemented for probabilistic \
                surrogate models."
            )

        if not self.surrogate_model.is_torch:
            raise AttributeError(
                "This BoTorch-based sampler is experimental and currently only "
                "supports GPyTorch GP surrogates"
            )

    def optimize_acqf_and_get_observation(self, acq_func):
        """
        Optimizes the acquisition function, and returns a new candidate and a
        noisy observation.
        """
        # optimize
        candidates, acq_values = optimize_acqf(
            acq_function=acq_func,
            bounds=self.bounds,
            q=self.q,
            num_restarts=self.num_restarts,
            raw_samples=self.raw_samples,  # used for intialization heuristic
            options=self.optimizer_options,
            return_best_only=True,
        )
        # observe new values
        new_x = candidates.detach().cpu().numpy()
        new_y = self.observer(new_x)
        return new_x, new_y

    def sample(
        self,
        n_initial_points: int = 20,
        n_new_points_per_iteration: int = 1,
        stopping_criterion: float = 0.05,
        max_n_iterations: int = 5000,
    ):

        # Initial sampling
        if (n_initial_points is None) and (
            (self.fit_points_x is None) or (self.fit_points_y is None)
        ):
            raise ValueError(
                "Either `n_initial_points` or a pair of samples and target values"
                "`fit_points_x` and `fit_points_y` must be specified"
            )

        if n_initial_points is None:
            n_initial_points = self.fit_points_x.shape[0]

        # Draw mc samples
        self.mc_samples = self.func_sample(self.n_mc_points)

        if stopping_criterion and not self.evaluation_metric:
            raise ValueError("Specified stopping criterion but no evaluation metric.")

        logger.info(
            f"qNIPV sampling for {max_n_iterations} max iterations with "
            f"{n_initial_points} initial points and stopping "
            f"criterion = {stopping_criterion}."
        )

        # If no initial points are passed, use latin hypercube sampling.
        if self.fit_points_x is None:

            # Draw samples
            self.fit_points_x = self.func_sample(
                n=n_initial_points,
            )

            # Evaluate
            self.fit_points_y = self.observer(self.fit_points_x)

        iteration = 0
        convergence = False
        while (convergence is False) and (iteration < max_n_iterations):

            start_time = time.time()

            # Initial fit of surrogate model
            self.surrogate_model.fit(self.fit_points_x, self.fit_points_y)
            logger.info(
                f"Fitted a new surrogate model in {time.time() - start_time} sec."
            )

            # Define acquisition function
            # Objective and sampler set to None based on this issue:
            # https://github.com/pytorch/botorch/issues/573
            qNIPV = qNegIntegratedPosteriorVariance(
                model=self.surrogate_model.model,
                mc_points=torch.tensor(self.mc_samples),
                # sampler=qmc_sampler,
                objective=None,
                sampler=None,
            )

            # Optimize acquisition function
            start_time = time.time()
            new_x, new_y = self.optimize_acqf_and_get_observation(qNIPV)
            logger.info(
                f"Finished optimizing acquisition function in "
                f"{time.time() - start_time} sec."
            )

            # Check user-specified stopping criterion
            score = self.metric(
                self.surrogate_model.predict(self.test_points_x), self.test_points_y
            )

            if stopping_criterion:
                if score <= stopping_criterion:
                    logger.info(f"Algorithm converged in {iteration} iterations")
                    convergence = True

            # Update training points
            self.fit_points_x = torch.cat([self.fit_points_x, new_x])
            self.fit_points_y = torch.cat(
                [
                    self.fit_points_y,
                    new_y.mreshape(
                        -1,
                    ),
                ]
            )

            iteration += 1

            if self.verbose:
                print(f"Sampling iteration = {iteration} / {max_n_iterations}")

        return self.fit_points_x, self.fit_points_y
